/* -*-java-*- */
%{
    #include "../dbxml/Value.hpp"
	#include "../dbxml/RawNodeValue.hpp"
	#include "../dbxml/Results.hpp"
	#include "../dbxml/Document.hpp"
	#include "../dbxml/ScopedDbt.hpp"

	using namespace DbXml;
    typedef struct
    {
	XmlDocument doc;
	XmlEventWriter *writer;
    } DocumentEventWriter;
    %}

%include "various.i"

    %module(directors = "1") dbxml_java

 // Enable the "director" feature for allowing Java
 // implementations of virtual behavior for 2 classes,
 // XmlInputStream and XmlResolver.
 // NOTE: there are issues getting it to work properly
 // with XmlInputStream, because of char * -> byte[]
 // mappings.  This will need to be fixed later.  For now,
 // it's not possible to implement XmlInputStream subclasses in Java
    %feature("director") XmlResolver;

%{
    class JavaException {};

    class XmlInputStream_ReadBytes_InOutByteArray {
	public:
	    XmlInputStream_ReadBytes_InOutByteArray(JNIEnv *jenv, jbyteArray &jtoFill, char *toFill, unsigned int &result)
	    : jenv_(jenv), jtoFill_(jtoFill), toFill_(toFill), result_(result)
	{
	}
	~XmlInputStream_ReadBytes_InOutByteArray() {
	    if(jenv_->ExceptionOccurred()) {
		// This is caught in the exception handling routines
		throw JavaException();
	    } else {
		jenv_->GetByteArrayRegion(jtoFill_, (jsize)0, (jsize)result_, (jbyte*)toFill_);
	    }
	}
	private:
	    JNIEnv *jenv_;
	jbyteArray &jtoFill_;
	char *toFill_;
	unsigned int &result_;
    };
    %}

// this turns the char * into byte[] for Java
%apply signed char *INOUT {char *toFill};
%typemap(freearg) char *toFill ""
    %typemap(directorin,descriptor="[B") char *toFill %{
    $input = jenv->NewByteArray((jsize)maxToRead);
    XmlInputStream_ReadBytes_InOutByteArray arg_guard$argnum(jenv, $input, $1, result);
    %}

%feature("director") XmlInputStream;


%include "enumtypeunsafe.swg"
    %javaconst(1);

// These two typemaps for string are directly copied from the std_string.i typemap
// in the java directory that comes with SWIG.  It is modified
// to map a null input into an empty string, rather than throwing
// an exception
%typemap(in) std::string
    %{ 
    const char *$1_pstr = "";
    if($input) {
	$1_pstr = (const char *)jenv->GetStringUTFChars($input, 0);
    }
    if (!$1_pstr) return $null;
    $1 =  std::string($1_pstr);
    if ($input) {
	jenv->ReleaseStringUTFChars($input, $1_pstr);
    }
    %}

%typemap(in) const std::string &
    %{ 
    const char *$1_pstr = "";
    if($input) {
	$1_pstr = (const char *)jenv->GetStringUTFChars($input, 0);
    }
    if (!$1_pstr) return $null;
    std::string $1_str($1_pstr);
    $1 = &$1_str;
    if ($input) {
	jenv->ReleaseStringUTFChars($input, $1_pstr);
    }
    %}

//
// Pile of typemaps for unsigned char to be equivalent
// to signed char (Java String).  Not all may be required, but
// experimentation is necessary to eliminate any.
// NOTE: I tried using just these two, but they resulted in
// no compilation due to missing some explicit casts to JNI calls
//  %apply char {unsigned char};
//  %apply char * {unsigned char*};
//
%typemap(jni) unsigned char, const unsigned char & "jchar"
     %typemap(jtype) unsigned char, const unsigned char & "char"
     %typemap(jstype) unsigned char, const unsigned char & "char"
     %typemap(jni) unsigned char *, unsigned char[ANY], unsigned char[]    "jstring"
     %typemap(jtype) unsigned char *, unsigned char[ANY], unsigned char[]  "String"
     %typemap(jstype) unsigned char *, unsigned char[ANY], unsigned char[] "String"
     %typemap(javain) unsigned char, const unsigned char & "$javainput"
     %typemap(javain) unsigned char *, unsigned char[ANY], unsigned char[] "$javainput"
     %typemap(javaout) unsigned char, const unsigned char & { return $jnicall; }
%typemap(javaout) unsigned char *, unsigned char[ANY], unsigned char[] {
    return $jnicall;
 }

%typemap(in, noblock=1) unsigned char * {
    $1 = 0;
    if ($input) {
	$1 = ($1_ltype)JCALL2(GetStringUTFChars, jenv, $input, 0);
	if (!$1) return $null;
    }
 }
%typemap(in, noblock=1) const unsigned char * {
    $1 = 0;
    if ($input) {
	$1 = ($1_ltype)JCALL2(GetStringUTFChars, jenv, $input, 0);
	if (!$1) return $null;
    }
 }

%typemap(out, noblock=1) const unsigned char * { if ($1) $result = JCALL1(NewStringUTF, jenv, (const char *) $1); }

%typemap(out, noblock=1) unsigned char * { if ($1) $result = JCALL1(NewStringUTF, jenv, (char *) $1); }

// end pile of typemaps for unsigned char

%exception {
    jthrowable t = NULL;
    try {
	$action
	    }
    catch (std::exception &se) {
	t = createException(jenv, &se);
    }
    catch(JavaException & /* je */) {
	// This means there's already an exception waiting in the JVM
	return $null;
    }
    catch (...) {
	t = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
					XmlException::INTERNAL_ERROR,
					jenv->NewStringUTF("Uncaught exception from C++ API"), 0, 0, 0, 0);
    }

    if (t) {
	jenv->Throw(t);
	return $null;
    }
 }

    %define JAVA_XMLEXCEPTION(_fname)
    %javaexception("XmlException") _fname {
    jthrowable t = NULL;
    try {
	$action
	    }
    catch (std::exception &se) {
	t = createException(jenv, &se);
    }
    catch(JavaException & /* je */) {
	// This means there's already an exception waiting in the JVM
	return $null;
    }
    catch (...) {
	t = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
					XmlException::INTERNAL_ERROR,
					jenv->NewStringUTF("Uncaught exception from C++ API"), 0, 0, 0, 0);
    }

    if (t) {
	jenv->Throw(t);
	return $null;
    }
 }
%enddef

    %typemap(javaimports) XmlTransaction %{
    import com.sleepycat.db.Transaction;
    import com.sleepycat.db.DatabaseException;
    %}

%typemap(javaimports) XmlManager %{
    import com.sleepycat.db.Environment;
    import com.sleepycat.db.Transaction;
    import com.sleepycat.db.TransactionConfig;
    import com.sleepycat.db.VerifyConfig;
    import com.sleepycat.db.DatabaseException;
    import com.sleepycat.db.internal.DbEnv;
    import com.sleepycat.db.internal.DbConstants;
    import com.sleepycat.db.XmlHelper;
    %}

%typemap(javaimports) XmlResults %{
    import java.util.*;
    %}

%pragma(java) jniclassimports=%{
     import com.sleepycat.db.DatabaseException;
     %}

%pragma(java) jniclasscode=%{
     static {
	 // An alternate library name can be specified via a property.
	 //
	 String override;

	 if ((override = System.getProperty("sleepycat.dbxml.libfile")) != null) {
	     System.load(override);
	 }
	 else if ((override = System.getProperty("sleepycat.dbxml.libname")) != null) {
	     System.loadLibrary(override);
	 }
	 else {
	     String os = System.getProperty("os.name");
	     if (os != null && os.startsWith("Windows")) {
		 // library name is "libdbxml_javaXX.dll" (for example) on Windows
		 String libname = "libdbxml_java" +
		     XmlConstants.DBXML_VERSION_MAJOR + XmlConstants.DBXML_VERSION_MINOR;

		 // first try to load the release version. If that fails try debug
		 try {
		     System.loadLibrary(libname);
		 } catch (UnsatisfiedLinkError e) {
		     try {
			 libname += "d";
			 System.loadLibrary(libname);
		     } catch (UnsatisfiedLinkError e2) {
			 throw e;
		     }
		 }
	     }
	     else {
		 // library name is "libdbxml_java-Major.Minor.so" (for example) on UNIX
		 // Note: "dbxml_java" isn't good enough;
		 // some Unixes require us to use the explicit SONAME.
		 System.loadLibrary("dbxml_java-" +
				    XmlConstants.DBXML_VERSION_MAJOR + "." +
				    XmlConstants.DBXML_VERSION_MINOR);
	     }
	 }

	 initialize();
     }

     static native final void initialize();
     public final static native long new_XmlResults(long jarg1, XmlResults jarg1_);
     public final static native void materializeMetaData(XmlDocument document);
     public final static native void interruptQuery(long queryContext);
     %}

%typemap(check) SWIGTYPE *self %{
    if (!$input) {
	jenv->Throw((jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						XmlException::INTERNAL_ERROR,
						jenv->NewStringUTF("null object - call after object destroyed?"),
						0, 0, 0, 0));					
	return $null;
    }
    %}

%typemap(check) SWIGTYPE *jcls %{
    UNUSED($input);	/* just in case */
    %}
    
%extend XmlManager{
    XmlDocument createDocumentInternal() {
	return self->createDocument();
    }
 }

/*
 * protect XmlManager methods that aren't publicly supported
 */
%javamethodmodifiers XmlManager::createTransaction(DbTxn*) "protected";
%javamethodmodifiers XmlManager::createContainer(
						 const std::string &name, 
						 u_int32_t flags,
						 enum XmlContainer::ContainerType type =
						 XmlContainer::NodeContainer,
						 int mode = 0) "protected";

%javamethodmodifiers XmlManager::createContainer(
						 XmlTransaction &txn,
						 const std::string &name, 
						 u_int32_t flags,
						 enum XmlContainer::ContainerType type =
						 XmlContainer::NodeContainer,
						 int mode = 0) "protected";

%javamethodmodifiers XmlManager::openContainer(const std::string &name, 
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlManager::openContainer(XmlTransaction &txn,
					       const std::string &name, 
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlManager::openContainer(XmlTransaction &txn,
					       const std::string &name, 
					       u_int32_t flags,
					       enum XmlContainer::ContainerType,
					       int mode = 0) "protected";

%javamethodmodifiers XmlManager::openContainer(const std::string &name, 
					       u_int32_t flags,
					       enum XmlContainer::ContainerType,
					       int mode = 0) "protected";

%javamethodmodifiers XmlManager::query(const std::string &query, 
 				       XmlQueryContext &context,
 				       u_int32_t flags) "protected";
 
%javamethodmodifiers XmlManager::query(XmlTransaction &txn,
 				       const std::string &query, 
 				       XmlQueryContext &context,
 				       u_int32_t flags) "protected";
 				       
%javamethodmodifiers XmlManager::createDocumentInternal() "protected";
						
%javamethodmodifiers XmlManager::verifyContainer(const std::string &name, 
						 const char *filename,
						 u_int32_t flags) "protected";

%javamethodmodifiers XmlManager::reindexContainer(const std::string &name, 
						  XmlUpdateContext &uc,
						  u_int32_t flags = 0) "protected";
%javamethodmodifiers XmlManager::reindexContainer(XmlTransaction &txn, 
						  const std::string &name,
						  XmlUpdateContext &uc, 
						  u_int32_t flags = 0) "protected";
%javamethodmodifiers XmlManager::createIndexLookupInternal(XmlContainer &cont,
							   const std::string &uri, 
							   const std::string &name,
							   const std::string &index,
							   const XmlValue &value = XmlValue(),
							   enum XmlIndexLookup::Operation op = XmlIndexLookup::NONE) "protected";


/*
 * The default typemap for delete makes it synchronized... change that
 */
%typemap(javadestruct, methodname="delete", methodmodifiers="public") SWIGTYPE {
    if(swigCPtr != 0 && swigCMemOwn) {
	swigCMemOwn = false;
	$jnicall;
    }
    swigCPtr = 0;
 }

/*
 * Pieces of XmlManager and XmlTransaction are implemented
 * directly in Java, for better integration with DB.
 * Also, some global constants are moved to the XmlManager class in Java
 */

/*
 * Re-map delete() for XmlManager to close the Environment if necessary,
 * using closeInternal() (protected).
 * The basic code here comes from java.swg
 */
%typemap(javadestruct, methodname="delete", methodmodifiers="public") XmlManager {
    closeInternal();
    if(swigCPtr != 0 && swigCMemOwn) {
	swigCMemOwn = false;
	$jnicall;
    }
    swigCPtr = 0;
 }
%typemap(javacode) XmlManager %{

    private Environment dbenv;
    private XmlContainerConfig defaultConfig;
    private boolean threaded = true; // default on if no Environment
    private boolean adopted = false;
    private XmlManagerConfig config = null;

    public XmlManager(final Environment dbenv,
		      XmlManagerConfig config)
	throws XmlException {
	this(dbxml_javaJNI.new_XmlManager(XmlHelper.getDbEnv(dbenv),
					  ((config == null ) ? 0 :
					   config.makeFlags())),true);
	this.dbenv = dbenv;
	this.config = config;
	if ((config != null) && (config.getAdoptEnvironment()))
	    this.adopted = true;
	try {
	    threaded =
		((XmlHelper.getDbEnv(dbenv).get_open_flags() &
		  DbConstants.DB_THREAD) != 0);
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
    }

    public XmlManager(XmlManagerConfig config)
	throws XmlException, java.io.FileNotFoundException {
	DbEnv newEnv = null;
	try {
	    newEnv = new DbEnv(0);
	    newEnv.set_cachesize(64 * 1024 * 1024, 1); // 64MB cache
	    newEnv.set_errpfx("BDB XML");
	    newEnv.open(null, DbConstants.DB_PRIVATE|
			DbConstants.DB_CREATE|DbConstants.DB_INIT_MPOOL|
			DbConstants.DB_THREAD, 0);
	    this.dbenv = XmlHelper.makeNewEnvironment(newEnv);
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	this.adopted = true;
	this.config = config;
	// do what the SWIG-generated ctor does, in order to
	// create a C++ XmlManager object.  Add ADOPT flag,
	// since the DbEnv was internally constructed.
	int flags = (config == null) ? 0 : config.makeFlags();
	swigCPtr = dbxml_javaJNI.new_XmlManager(newEnv, flags);
	swigCMemOwn = true;
    }

    public XmlManager()
	throws XmlException, java.io.FileNotFoundException {
	this(null);
    }

    public Environment getEnvironment() {
	return dbenv;
    }

    public XmlManagerConfig getManagerConfig() {
	return config;
    }

    public void setDefaultContainerConfig(XmlContainerConfig config) {
	this.defaultConfig = config;
    }

    public XmlContainerConfig getDefaultContainerConfig() {
	return defaultConfig;
    }
    
    public XmlUpdateContext createUpdateContext() { 
	return new XmlUpdateContext(this);
    }
    
    public XmlQueryContext createQueryContext() throws XmlException{
	XmlQueryContext xqc = 
	    new XmlQueryContext(XmlQueryContext.LiveValues, XmlQueryContext.Eager);
	xqc.mgr = this;
	return xqc;
    }
    
    public XmlQueryContext createQueryContext(int rt, int et) throws XmlException{
	XmlQueryContext xqc = new XmlQueryContext(rt, et);
	xqc.mgr = this;
	return xqc;
    }
    
    public XmlQueryContext createQueryContext(int rt) throws XmlException{
	XmlQueryContext xqc = new XmlQueryContext(rt, XmlQueryContext.Eager);
	xqc.mgr = this;
	return xqc;
    }

    public XmlInputStream createInputStream(java.io.InputStream is) {
	return new NativeXmlInputStream(is);
    }

    public XmlContainer createContainer(String name)
	throws XmlException {
	return createContainer(name, defaultConfig);
    }

    public XmlContainer createContainer(XmlTransaction txn, String name)
	throws XmlException {
	return createContainer(txn, name, defaultConfig);
    }

    public XmlContainer openContainer(String name)
	throws XmlException {
	return openContainer(name, defaultConfig);
    }

    public XmlContainer openContainer(XmlTransaction txn, String name)
	throws XmlException {
	return openContainer(txn, name, defaultConfig);
    }

    public XmlContainer createContainer(String name,
					XmlContainerConfig config)
	throws XmlException {
	int containerType;
	int flags = 0;
	if (config != null) {
	    flags = config.makeFlags(threaded);
	    containerType = (config.getNodeContainer()) ?
		XmlContainer.NodeContainer : XmlContainer.WholedocContainer;
	} else {
	    containerType = getDefaultContainerType();
	}
	XmlContainer cont = createContainer(name, flags, containerType, 0);
	cont.setContainerConfig(config);
	cont.setManager(this);
	return cont;
    }

    public XmlContainer createContainer(XmlTransaction txn,
					String name,
					XmlContainerConfig config)
	throws XmlException {
	int flags = 0;
	int containerType;
	if (config != null) {
	    flags = config.makeFlags(threaded);
	    containerType = (config.getNodeContainer()) ?
		XmlContainer.NodeContainer : XmlContainer.WholedocContainer;
	} else {
	    containerType = getDefaultContainerType();
	}
	XmlContainer cont = createContainer(txn, name, flags, containerType, 0);
	cont.setContainerConfig(config);
	cont.setManager(this);
	return cont;
    }

    public XmlContainer openContainer(String name,
				      XmlContainerConfig config)
	throws XmlException {
	int flags = 0;
	int containerType;
	if (config != null) {
	    flags = config.makeFlags(threaded);
	    containerType = (config.getNodeContainer()) ?
		XmlContainer.NodeContainer : XmlContainer.WholedocContainer;
	} else {
	    containerType = getDefaultContainerType();
	}

	XmlContainer cont = openContainer(name, flags, containerType, 0);
	cont.setContainerConfig(config);
	cont.setManager(this);
	return cont;
    }

    public XmlContainer openContainer(XmlTransaction txn,
				      String name,
				      XmlContainerConfig config)
	throws XmlException {
	int flags = 0;
	int containerType;
	if (config != null) {
	    flags = config.makeFlags(threaded);
	    containerType = (config.getNodeContainer()) ?
		XmlContainer.NodeContainer : XmlContainer.WholedocContainer;
	} else {
	    containerType = getDefaultContainerType();
	}

	XmlContainer cont = openContainer(txn, name, flags, containerType, 0);
	cont.setContainerConfig(config);
	cont.setManager(this);
	return cont;
    }

    public XmlTransaction createTransaction(com.sleepycat.db.Transaction toUse)
	throws XmlException {
	XmlTransaction txn = createTransaction(XmlHelper.getDbTxn(toUse));
	txn.setTransaction(toUse);
	return txn;
    }

    public XmlTransaction createTransaction(final Transaction parent,
					    TransactionConfig config)
	throws XmlException {
	Transaction newTxn = null;
	try {
	    newTxn = dbenv.beginTransaction(parent, config);
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	return createTransaction(newTxn);
    }

    public XmlTransaction createTransaction()
	throws XmlException {
	return createTransaction(null, null);
    }
    
    public XmlIndexLookup createIndexLookup(XmlContainer cont, 
					    String uri, String name, String index, 
					    XmlValue value, int op) 
	throws XmlException {
	XmlIndexLookup lookup = createIndexLookupInternal(cont, uri, name, index, value, op);
	lookup.container = cont;
	lookup.lowBoundValue = value;
	return lookup;
    }
	
    public XmlIndexLookup createIndexLookup(XmlContainer cont, 
					    String uri, String name, String index, 
					    XmlValue value) 
	throws XmlException {
	XmlIndexLookup lookup = createIndexLookupInternal(cont, uri, name, index, value);
	lookup.container = cont;
	lookup.lowBoundValue = value;
	return lookup;
    }
	
    public XmlIndexLookup createIndexLookup(XmlContainer cont, 
					    String uri, String name, String index) 
	throws XmlException {
	XmlIndexLookup lookup = createIndexLookupInternal(cont, uri, name, index);
	lookup.container = cont;
	return lookup;
    }
    
    public XmlDocument createDocument()
	throws XmlException {
	XmlDocument doc = createDocumentInternal();
	doc.setManager(this);
	return doc;
    }

    public XmlResults query(String query,
			    XmlQueryContext context,
			    XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return query(query, context, flags);
    }

    public XmlResults query(XmlTransaction txn,
			    String query,
			    XmlQueryContext context,
			    XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return query(txn, query, context, flags);
    }

    public void reindexContainer(String name,
				 XmlContainerConfig config)
	throws XmlException {
	reindexContainer(name, new XmlUpdateContext(this), config);
    }
    
    public void reindexContainer(String name,
				 XmlUpdateContext context,
				 XmlContainerConfig config)
	throws XmlException {
	int flags = 0;
	if (config != null)
	    flags = config.makeFlags(threaded);
	reindexContainer(name, context, flags);
    }
    
    public void reindexContainer(XmlTransaction txn,
				 String name,
				 XmlContainerConfig config)
	throws XmlException {
	reindexContainer(txn, name, new XmlUpdateContext(this), config);
    }

    public void reindexContainer(XmlTransaction txn,
				 String name,
				 XmlUpdateContext context,
				 XmlContainerConfig config)
	throws XmlException {
	int flags = 0;
	if (config != null)
	    flags = config.makeFlags(threaded);
	reindexContainer(txn, name, context, flags);
    }

    public void verifyContainer(String name,
				String filename,
				VerifyConfig config)
	throws XmlException {
	int flags = 0;
	if (config.getAggressive())
	    flags |= DbConstants.DB_AGGRESSIVE;
	if (config.getSalvage())
	    flags |= DbConstants.DB_SALVAGE;
	if (config.getPrintable())
	    flags |= DbConstants.DB_PRINTABLE;
	if (config.getOrderCheckOnly())
	    flags |= DbConstants.DB_ORDERCHKONLY;
	if (config.getNoOrderCheck())
	    flags |= DbConstants.DB_NOORDERCHK;
	verifyContainer(name, filename, flags);
    }

    protected void closeInternal() {
	if (adopted) {
	    try {
		dbenv.close();
	    } catch (DatabaseException de) {}
	}
	adopted = false;
	this.dbenv = null;
	this.config = null;
    }

    public void close() {
	delete();
    }

    public final static int LEVEL_NONE = dbxml_java.LEVEL_NONE;
    public final static int LEVEL_DEBUG = dbxml_java.LEVEL_DEBUG;
    public final static int LEVEL_INFO = dbxml_java.LEVEL_INFO;
    public final static int LEVEL_WARNING = dbxml_java.LEVEL_WARNING;
    public final static int LEVEL_ERROR = dbxml_java.LEVEL_ERROR;
    public final static int LEVEL_ALL = dbxml_java.LEVEL_ALL;
    public final static int CATEGORY_NONE = dbxml_java.CATEGORY_NONE;
    public final static int CATEGORY_INDEXER = dbxml_java.CATEGORY_INDEXER;
    public final static int CATEGORY_QUERY = dbxml_java.CATEGORY_QUERY;
    public final static int CATEGORY_OPTIMIZER = dbxml_java.CATEGORY_OPTIMIZER;
    public final static int CATEGORY_DICTIONARY = dbxml_java.CATEGORY_DICTIONARY;
    public final static int CATEGORY_CONTAINER = dbxml_java.CATEGORY_CONTAINER;
    public final static int CATEGORY_NODESTORE = dbxml_java.CATEGORY_NODESTORE;
    public final static int CATEGORY_MANAGER = dbxml_java.CATEGORY_MANAGER;
    public final static int CATEGORY_ALL = dbxml_java.CATEGORY_ALL;
    public final static String metaDataNamespace_uri = dbxml_javaConstants.metaDataNamespace_uri;
    public final static String metaDataNamespace_prefix = dbxml_javaConstants.metaDataNamespace_prefix;
    public final static String metaDataName_name = dbxml_javaConstants.metaDataName_name;
    public final static String metaDataName_root = dbxml_javaConstants.metaDataName_root;
    %}


/*
 * XmlTransaction is mostly Java
 */
%typemap(javacode) XmlTransaction %{
    Transaction txn;

    public void setTransaction(final com.sleepycat.db.Transaction txn) {
	this.txn = txn;
    }

    public Transaction getTransaction() {
	return this.txn;
    }

    public void abort()
	throws XmlException {
	try {
	    txn.abort();
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	delete();
    }

    public void commit()
	throws XmlException {
	try {
	    txn.commit();
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	delete();
    }

    public void commitSync()
	throws XmlException {
	try {
	    txn.commitSync();
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	delete();
    }

    public void commitNoSync()
	throws XmlException {
	try {
	    txn.commitNoSync();
	} catch (DatabaseException de) {
	    throw new XmlException(XmlException.DATABASE_ERROR,
				   de.toString(), de,
				   de.getErrno(), 0, 0);
	}
	delete();
    }
    %}

/*
 * protect XmlContainer methods that aren't publicly supported
 */
JAVA_XMLEXCEPTION(putDocument)
    %javamethodmodifiers XmlContainer::putDocument(XmlDocument &,
						   XmlUpdateContext &,
						   u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(XmlTransaction &,
					       XmlDocument &,
					       XmlUpdateContext &,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(const std::string &name,
					       XmlInputStream *input,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(XmlTransaction &,
					       const std::string &name,
					       XmlInputStream *input,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(XmlTransaction &,
					       const std::string &name, 
					       XmlEventReader &reader,
					       XmlUpdateContext &context, 
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(const std::string &name, 
					       XmlEventReader &reader,
					       XmlUpdateContext &context, 
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(const std::string &name,
					       const std::string &contents,
					       XmlUpdateContext &context,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocument(XmlTransaction &,
					       const std::string &name,
					       const std::string &contents,
					       XmlUpdateContext &context,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocumentAsEventWriter(XmlTransaction &,
							    XmlDocument &document,
							    u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::putDocumentAsEventWriter(XmlDocument &document,
							    u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::getAllDocuments(u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::getAllDocuments(XmlTransaction &,
						   u_int32_t flags) "protected";						   
%javamethodmodifiers XmlContainer::getNode(XmlTransaction&, 
					   const std::string&, u_int32_t) "protected";						
%javamethodmodifiers XmlContainer::getIndexSpecification(XmlTransaction &,
							 u_int32_t flags) const "protected";							 

%javamethodmodifiers XmlContainer::lookupIndex(
					       XmlQueryContext &context, 
					       const std::string &uri,
					       const std::string &name, 
					       const std::string &index,
					       const XmlValue &value,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::lookupIndex(
					       XmlQueryContext &context, 
					       const std::string &uri,
					       const std::string &name, 
					       const std::string &parent_uri,
					       const std::string &parent_name, 
					       const std::string &index,
					       const XmlValue &value,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::lookupIndex(XmlTransaction &,
					       XmlQueryContext &context, 
					       const std::string &uri,
					       const std::string &name, 
					       const std::string &index,
					       const XmlValue &value,
					       u_int32_t flags) "protected";

%javamethodmodifiers XmlContainer::lookupIndex(XmlTransaction &,
					       XmlQueryContext &context, 
					       const std::string &uri,
					       const std::string &name, 
					       const std::string &parent_uri,
					       const std::string &parent_name, 
					       const std::string &index,
					       const XmlValue &value,
					       u_int32_t flags) "protected";
%javamethodmodifiers XmlContainer::lookupIndex(XmlTransaction &, XmlQueryContext &,
					       const std::string &, const std::string &,
					       const std::string &, const XmlValue &,
					       u_int32_t)"protected";

%javamethodmodifiers XmlContainer::getDocumentInternal(const std::string &, u_int32_t) "protected";

%javamethodmodifiers XmlContainer::getDocumentInternal(XmlTransaction &, const std::string &, u_int32_t) "protected";

%javamethodmodifiers XmlContainer::putDocumentInternal(XmlDocument &,u_int32_t) "protected";
%javamethodmodifiers XmlContainer::putDocumentInternal(XmlTransaction &, XmlDocument &,u_int32_t) "protected";
%javamethodmodifiers XmlContainer::deleteDocumentInternal(XmlDocument &) "protected";
%javamethodmodifiers XmlContainer::deleteDocumentInternal(XmlTransaction &, XmlDocument &) "protected";
%javamethodmodifiers XmlContainer::updateDocumentInternal(XmlDocument &) "protected";
%javamethodmodifiers XmlContainer::updateDocumentInternal(XmlTransaction &, XmlDocument &) "protected";	

%extend XmlContainer{
    XmlDocument getDocumentInternal(const std::string &name, u_int32_t flags)
	{
		return self->getDocument(name, flags);
	}

    XmlDocument getDocumentInternal(XmlTransaction &txn, const std::string &name, u_int32_t flags)
	{
		return self->getDocument(txn, name, flags);
	}

    XmlDocument &putDocumentInternal(XmlDocument &document,
				     u_int32_t flags = 0)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->putDocument(document, context, flags);
	    return document;
	}
    XmlDocument &putDocumentInternal(XmlTransaction &txn, XmlDocument &document,
				     u_int32_t flags = 0)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->putDocument(txn, document, context, flags);
	    return document;
	}
			 
    DocumentEventWriter putDocumentAsEventWriter(XmlDocument &document,
						 u_int32_t flags)
    {
	DocumentEventWriter dew;
	XmlUpdateContext context = self->getManager().createUpdateContext();
	dew.writer = &self->putDocumentAsEventWriter(document, context, flags);
	dew.doc = document;
	return dew;
    }
    DocumentEventWriter putDocumentAsEventWriter(XmlTransaction &txn,
						 XmlDocument &document,
						 u_int32_t flags)
    {
	DocumentEventWriter dew;
	XmlUpdateContext context = self->getManager().createUpdateContext();
	dew.writer = &self->putDocumentAsEventWriter(txn, document, context, flags);
	dew.doc = document;
	return dew;
    }
						 
    XmlDocument &deleteDocumentInternal(XmlDocument &document)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->deleteDocument(document, context);
	    return document;
	}
    XmlDocument &deleteDocumentInternal(XmlTransaction &txn, XmlDocument &document)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->deleteDocument(txn, document, context);
	    return document;
	}
	
    XmlDocument &updateDocumentInternal(XmlDocument &document)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->updateDocument(document, context);
	    return document;
	}
    XmlDocument &updateDocumentInternal(XmlTransaction &txn, XmlDocument &document)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    self->updateDocument(txn, document, context);
	    return document;
	}

    XmlResults getNode(const std::string &name) {
	XmlValue value = self->getNode(name, 0);
	XmlResults result = self->getManager().createResults();
	result.add(value);
	return result;
    }
    
    XmlResults getNode(XmlTransaction &txn, const std::string &name,
		       u_int32_t flags) {
	XmlValue value = self->getNode(txn, name, flags);
	XmlResults result = new ValueResults(self->getManager(), txn);
	result.add(value);
	return result;
    }
    
    std::string putDocument(const std::string &name, XmlInputStream *input,
			    u_int32_t flags)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    return self->putDocument(name, input, context, flags);
	}

    std::string putDocument(XmlTransaction &txn, const std::string &name,
			    XmlInputStream *input,
			    u_int32_t flags)
	{
	    XmlUpdateContext context = self->getManager().createUpdateContext();
	    return self->putDocument(txn, name, input, context, flags);
	}
}

%typemap(javacode) XmlContainer %{
    private XmlContainerConfig config;
    private XmlManager manager;

    public XmlContainerConfig getContainerConfig() {
	return config;
    }

    public XmlManager getManager() {
	return manager;
    }

    public void close() {
	delete(); // delete the C++ object, as well
    }
    
    // Start XmlDocument-based putDocument methods
    public void putDocument(XmlDocument document)
	throws XmlException {
	putDocument(null, document, null, null);
    }
    
    public void putDocument(XmlDocument document,
			    XmlUpdateContext context)
	throws XmlException {
	putDocument(null, document, null, null);
    }
    
    public void putDocument(XmlDocument document,
			    XmlDocumentConfig config)
	throws XmlException {
	putDocument(null, document, null, config);
    }

    public void putDocument(XmlDocument document,
			    XmlUpdateContext context,
			    XmlDocumentConfig config)
	throws XmlException {
	putDocument(null, document, null, config);
    }
    
    public void putDocument(XmlTransaction txn, XmlDocument document)
	throws XmlException {
	putDocument(txn, document, null, null);
    }
    
    public void putDocument(XmlTransaction txn, XmlDocument document,
			    XmlUpdateContext context)
	throws XmlException {
	putDocument(txn, document, null, null);
    }
    
    public void putDocument(XmlTransaction txn, XmlDocument document,
			    XmlDocumentConfig config)
	throws XmlException {
	putDocument(txn, document, null, config);
    }

    // all putDocument calls funnel through this one
    public void putDocument(XmlTransaction txn, XmlDocument document,
			    XmlUpdateContext context,
			    XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	XmlDocument doc;
	if (txn != null)
	    doc = putDocumentInternal(txn, document, flags);
	else
	    doc = putDocumentInternal(document, flags);
	document.copy(doc);
	document.setManager(manager);
	if (txn != null)
	    document.setTransaction(txn);
    }
    
    // Start XmlInputStream-based putDocument methods
    public String putDocument(String name,
			      XmlInputStream input)
	throws XmlException {
	return putDocument(null, name, input, null, null);
    }
    
    public String putDocument(String name,
			      XmlInputStream input,
			      XmlUpdateContext context)
	throws XmlException {
	return putDocument(null, name, input, null, null);
    }
    
    public String putDocument(String name,
			      XmlInputStream input,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, input, null, config);
    }

    public String putDocument(String name,
			      XmlInputStream input,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, input, null, config);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlInputStream input)
	throws XmlException {
	return putDocument(txn, name, input, null, null);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlInputStream input,
			      XmlUpdateContext context)
	throws XmlException {
	return putDocument(txn, name, input, null, null);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlInputStream input,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(txn, name, input, null, config);
    }

    // XmlInputStream putDocument methods funnel through here
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlInputStream input,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	String docName = "";
	try{
	    if (txn != null)
		docName = putDocument(txn, name, input, flags);
	    else
		docName = putDocument(name, input, flags);
	}catch(XmlException e){
	    input.release();
	    throw e;
	}
	input.release();
	return docName;
    }
    
    // Start String-based putDocument methods
    public String putDocument(String name,
			      String content)
	throws XmlException {
	return putDocument(null, name, content, null, null);
    }
    
    public String putDocument(String name,
			      String content,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, content, null, config);
    }

    public String putDocument(String name,
			      String content,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, content, null, config);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      String content)
	throws XmlException {
	return putDocument(txn, name, content, null, null);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      String content,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(txn, name, content, null, config);
    }

    // String-based putDocument calls funnel through here
    public String putDocument(XmlTransaction txn,
			      String name,
			      String content,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	if (context == null)
	    context = new XmlUpdateContext(manager);
	if (txn != null)
	    return putDocument(txn, name, content, context, flags);
	else
	    return putDocument(name, content, context, flags);
    }

    // Start XmlEventReader-based putDocument methods    
    public String putDocument(String name,
			      XmlEventReader reader)
	throws XmlException {
	return putDocument(null, name, reader, null, null);
    }
    
    public String putDocument(String name,
			      XmlEventReader reader,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, reader, null, config);
    }

    public String putDocument(String name,
			      XmlEventReader reader,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(null, name, reader, context, config);
    }

    public String putDocument(String name,
			      XmlEventReader reader,
			      XmlUpdateContext context)
	throws XmlException {
	return putDocument(null, name, reader, context, null);
    }
    
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlEventReader reader,
			      XmlDocumentConfig config)
	throws XmlException {
	return putDocument(txn, name, reader, null, config);
    }

    public String putDocument(XmlTransaction txn, 
			      String name,
			      XmlEventReader reader)
	throws XmlException {
	return putDocument(txn, name, reader, null, null);
    }

    public String putDocument(XmlTransaction txn, 
			      String name,
			      XmlEventReader reader,
			      XmlUpdateContext context)
	throws XmlException {
	return putDocument(txn, name, reader, context, null);
    }

    // all XmlEventReader-based putDocument methods funnel here
    public String putDocument(XmlTransaction txn,
			      String name,
			      XmlEventReader reader,
			      XmlUpdateContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	String ret = "";
	if (context == null)
	    context = new XmlUpdateContext(manager);
	try{
	    if (txn != null)
		ret = putDocument(txn, name, reader, context, flags);
	    else
		ret = putDocument(name, reader, context, flags);
	}catch(XmlException e){
	    reader.release();
	    throw e;
	}
	reader.release(); /* no further access */
	return ret;
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlDocument document)
	throws XmlException {
	return putDocumentAsEventWriter(null, document, null, null);
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlDocument document,
						   XmlDocumentConfig config)
	throws XmlException {
	return putDocumentAsEventWriter(null, document, null, config);
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlDocument document,
						   XmlUpdateContext context)
	throws XmlException {
	return putDocumentAsEventWriter(null, document, null, null);
    }

    public XmlEventWriter putDocumentAsEventWriter(XmlDocument document,
						   XmlUpdateContext context,
						   XmlDocumentConfig config)
	throws XmlException {
	return putDocumentAsEventWriter(null, document, null, config);
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlTransaction txn,
						   XmlDocument document)
	throws XmlException {
	return putDocumentAsEventWriter(txn, document, null, null);
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlTransaction txn,
						   XmlDocument document,
						   XmlUpdateContext context)
	throws XmlException {
	return putDocumentAsEventWriter(txn, document, null, null);
    }
    
    public XmlEventWriter putDocumentAsEventWriter(XmlTransaction txn,
						   XmlDocument document,
						   XmlDocumentConfig config)
	throws XmlException {
	return putDocumentAsEventWriter(txn, document, null, config);
    }

    // All calls to putDocumentAsEventWriter funnel through this one
    public XmlEventWriter putDocumentAsEventWriter(XmlTransaction txn,
						   XmlDocument document,
						   XmlUpdateContext context,
						   XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	document.clearContent(); // this call clears any existing content
	XmlDocument doc;
	if (txn != null)
	    doc = (XmlDocument)putDocumentAsEventWriter(txn, document, flags);
	else
	    doc = (XmlDocument)putDocumentAsEventWriter(document, flags);
	document.copy(doc);
	document.setManager(manager);
	if (txn != null)
	    document.setTransaction(txn);
	long cPtr = doc.getEventWriterPtr();
	return (cPtr == 0) ? null : new XmlEventWriter(cPtr, true);
    }
    
    public void deleteDocument(String name)
	throws XmlException {
	deleteDocument(name, null);
    }
    
    public void deleteDocument(XmlDocument document)
	throws XmlException {
	deleteDocument(document, null);
    }
    
    public void deleteDocument(XmlDocument document, XmlUpdateContext context)
	throws XmlException {
	XmlDocument doc = deleteDocumentInternal(document);
	document.copy(doc);
	document.clearID();
	document.setManager(manager);
    }
    
    public void deleteDocument(XmlTransaction txn, String name)
	throws XmlException {
	deleteDocument(txn, name, null);
    }
    
    public void deleteDocument(XmlTransaction txn, XmlDocument document)
	throws XmlException {
	deleteDocument(txn, document, null);
    }
    
    public void deleteDocument(XmlTransaction txn, XmlDocument document,
			       XmlUpdateContext context)
	throws XmlException {
	XmlDocument doc = deleteDocumentInternal(txn, document);
	document.copy(doc);
	document.clearID();
	document.setManager(manager);
	document.setTransaction(txn);
    }
    
    public void updateDocument(XmlDocument document)
	throws XmlException {
	updateDocument(document, null);
    }
	
    public void updateDocument(XmlDocument document,
			       XmlUpdateContext context)
	throws XmlException {
	XmlDocument doc = updateDocumentInternal(document);
	document.copy(doc);
	document.setManager(manager);
    }
    
    public void updateDocument(XmlTransaction txn, XmlDocument document)
	throws XmlException {
	updateDocument(txn, document, null);
    }
    
    public void updateDocument(XmlTransaction txn, XmlDocument document,
			       XmlUpdateContext context)
	throws XmlException {
	XmlDocument doc = updateDocumentInternal(txn, document);
	document.copy(doc);
	document.setManager(manager);
	document.setTransaction(txn);
    }
    
    public void setIndexSpecification(XmlIndexSpecification index) throws XmlException {
	setIndexSpecification(index, new XmlUpdateContext(manager));
    }

    public void addIndex(String uri, String name, String index) throws XmlException {
	addIndex(uri, name, index, new XmlUpdateContext(manager));
    }

    public void addIndex(String uri, String name, int indexType, int syntaxType) throws XmlException {
	addIndex(uri, name, indexType, syntaxType, new XmlUpdateContext(manager));
    }

    public void deleteIndex(String uri, String name, String index) throws XmlException {
	deleteIndex(uri, name, index, new XmlUpdateContext(manager));
    }

    public void replaceIndex(String uri, String name, String index) throws XmlException {
	replaceIndex(uri, name, index, new XmlUpdateContext(manager));
    }

    public void addDefaultIndex(String index) throws XmlException {
	addDefaultIndex(index, new XmlUpdateContext(manager));
    }

    public void deleteDefaultIndex(String index) throws XmlException {
	deleteDefaultIndex(index, new XmlUpdateContext(manager));
    }

    public void replaceDefaultIndex(String index) throws XmlException {
	replaceDefaultIndex(index, new XmlUpdateContext(manager));
    }

    public void setIndexSpecification(XmlTransaction txn, XmlIndexSpecification index) throws XmlException {
	setIndexSpecification(txn, index, new XmlUpdateContext(manager));
    }

    public void addIndex(XmlTransaction txn, String uri, String name, String index) throws XmlException {
	addIndex(txn, uri, name, index, new XmlUpdateContext(manager));
    }

    public void addIndex(XmlTransaction txn, String uri, String name, int indexType, int syntaxType) throws XmlException {
	addIndex(txn, uri, name, indexType, syntaxType, new XmlUpdateContext(manager));
    }

    public void deleteIndex(XmlTransaction txn, String uri, String name, String index) throws XmlException {
	deleteIndex(txn, uri, name, index, new XmlUpdateContext(manager));
    }

    public void replaceIndex(XmlTransaction txn, String uri, String name, String index) throws XmlException {
	replaceIndex(txn, uri, name, index, new XmlUpdateContext(manager));
    }

    public void addDefaultIndex(XmlTransaction txn, String index) throws XmlException {
	addDefaultIndex(txn, index, new XmlUpdateContext(manager));
    }

    public void deleteDefaultIndex(XmlTransaction txn, String index) throws XmlException {
	deleteDefaultIndex(txn, index, new XmlUpdateContext(manager));
    }

    public void replaceDefaultIndex(XmlTransaction txn, String index) throws XmlException {
	replaceDefaultIndex(txn, index, new XmlUpdateContext(manager));
    }

    public XmlDocument getDocument(String name)
	throws XmlException {
	return getDocument(name, null);
    }

    public XmlDocument getDocument(String name,
				   XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	XmlDocument doc = getDocumentInternal(name, flags);
	doc.setManager(manager);
	return doc;
    }

    public XmlDocument getDocument(XmlTransaction txn,
				   String name)
	throws XmlException {
	return getDocument(txn, name, null);
    }

    public XmlDocument getDocument(XmlTransaction txn,
				   String name,
				   XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	XmlDocument doc = getDocumentInternal(txn, name, flags);
	doc.setManager(manager);
	doc.setTransaction(txn);
	return doc;
    }

    public XmlResults getAllDocuments(XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return getAllDocuments(flags);
    }


    public XmlResults getAllDocuments(XmlTransaction txn,
				      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return getAllDocuments(txn, flags);
    }
    
    public XmlResults getNode(XmlTransaction txn, String name, 
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return getNode(txn, name, flags);
    }
	
    public XmlIndexSpecification getIndexSpecification(XmlTransaction txn,
						       XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return getIndexSpecification(txn, flags);
    }

    public XmlResults lookupIndex(XmlQueryContext context,
				  String uri, String name,
				  String index,
				  XmlValue value,
				  XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return lookupIndex(context, uri, name, index, value, flags);
    }

    public XmlResults lookupIndex(XmlTransaction txn,
				  XmlQueryContext context,
				  String uri, String name,
				  String index,
				  XmlValue value,
				  XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return lookupIndex(txn, context, uri, name, index, value, flags);
    }

    public XmlResults lookupIndex(XmlQueryContext context,
				  String uri, String name,
				  String parentUri, String parentName,
				  String index,
				  XmlValue value,
				  XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return lookupIndex(context, uri, name,
			   parentUri, parentName, index, value, flags);
    }

    public XmlResults lookupIndex(XmlTransaction txn,
				  XmlQueryContext context,
				  String uri, String name,
				  String parentUri, String parentName,
				  String index,
				  XmlValue value,
				  XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return lookupIndex(txn, context, uri, name,
			   parentUri, parentName, index, value, flags);
    }

    /* package */
    void setContainerConfig(XmlContainerConfig config) {
	this.config = config;
    }

    /* package */
    void setManager(XmlManager manager) {
	this.manager = manager;
    }

    %}

/*
 * protect XmlQueryExpression methods that aren't publicly supported
 */

%javamethodmodifiers XmlQueryExpression::execute(XmlQueryContext &,
						 u_int32_t) const "protected";

%javamethodmodifiers XmlQueryExpression::execute(const XmlValue &, 
						 XmlQueryContext &,
						 u_int32_t) const "protected";

%javamethodmodifiers XmlQueryExpression::execute(XmlTransaction &,
						 XmlQueryContext &,
						 u_int32_t) const "protected";

%javamethodmodifiers XmlQueryExpression::execute(XmlTransaction &,
						 const XmlValue &, 
						 XmlQueryContext &,
						 u_int32_t) const "protected";

%typemap(javacode) XmlQueryExpression %{
    public XmlResults execute(XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(context, flags);
    }
    public XmlResults execute(XmlValue contextItem,
			      XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(contextItem, context, flags);
    }

    public XmlResults execute(XmlTransaction txn,
			      XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(txn, context, flags);
    }

    public XmlResults execute(XmlTransaction txn,
			      XmlValue contextItem,
			      XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(txn, contextItem, context, flags);
    }
    %}

%extend XmlIndexLookup{
    static XmlIndexLookup *copy(const XmlIndexLookup &o)
    {
	return new XmlIndexLookup(o);
    }
}
// protect non-exported XmlIndexLookup methods
%javamethodmodifiers XmlIndexLookup::setContainerInternal(XmlContainer &) "protected";
%javamethodmodifiers XmlIndexLookup::execute(XmlQueryContext &,
					     u_int32_t) const "protected";
%javamethodmodifiers XmlIndexLookup::execute(XmlTransaction &,
					     XmlQueryContext &,
					     u_int32_t) const "protected";
%javamethodmodifiers XmlIndexLookup::setLowBound(const XmlValue &,
						 enum XmlIndexLookup::Operation) "protected";
%javamethodmodifiers XmlIndexLookup::setHighBound(const XmlValue &,
						  enum XmlIndexLookup::Operation) "protected";
%javamethodmodifiers XmlIndexLookup::copy(const XmlIndexLookup&) "protected";

%typemap(javacode) XmlIndexLookup %{
    protected XmlContainer container;
    protected XmlValue lowBoundValue = new XmlValue();
    protected XmlValue highBoundValue = new XmlValue();
	
    public XmlIndexLookup(XmlIndexLookup o) throws XmlException {
	this(dbxml_javaJNI.XmlIndexLookup_copy(XmlIndexLookup.getCPtr(o), o), true);
	lowBoundValue = o.lowBoundValue;
	highBoundValue = o.highBoundValue;
    }
	
    public XmlContainer getContainer() 
	throws XmlException {
	return container;
    }
    public void setContainer(XmlContainer container)
	throws XmlException {
	this.container = container;
	setContainerInternal(container);
    }
    public XmlResults execute(XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(context, flags);
    }
    public XmlResults execute(XmlTransaction txn,
			      XmlQueryContext context,
			      XmlDocumentConfig config)
	throws XmlException {
	int flags = (config == null) ? 0 : config.makeFlags();
	return execute(txn, context, flags);
    }
    public XmlValue getLowBoundValue()
	throws XmlException {
	return lowBoundValue;
    }
    public XmlValue getHighBoundValue()
	throws XmlException {
	return highBoundValue;
    }
    public void setHighBound(XmlValue value, int op) throws XmlException {
	setHighBoundInternal(value,op);
	highBoundValue = value;
    }
    public void setLowBound(XmlValue value, int op) throws XmlException {
	setLowBoundInternal(value,op);
	lowBoundValue = value;
    }
    %}

%typemap(javadestruct, methodname="delete", methodmodifiers="protected") XmlEventReader
{
    if(swigCPtr != 0 && swigCMemOwn) {
	swigCMemOwn = false;
	$jnicall;
    }
    swigCPtr = 0;
}

%javamethodmodifiers XmlEventReader::closeInternal() "protected";

%typemap(javacode) XmlEventReader %{
    public void close()
	/* wrapper for close that protects from double-close */
	throws XmlException {
	if (swigCPtr != 0) {
	    closeInternal();
	    swigCPtr = 0;
	}
    }
    protected void release()
    {
	swigCMemOwn = false;
	swigCPtr = 0;
    }
    %}

%typemap(javadestruct, methodname="delete", methodmodifiers="protected") XmlEventWriter
{
    if(swigCPtr != 0 && swigCMemOwn) {
	swigCMemOwn = false;
	$jnicall;
    }
    swigCPtr = 0;
}

%javamethodmodifiers XmlEventWriter::closeInternal() "protected";

%typemap(javacode) XmlEventWriter %{
    public void close()
	/* wrapper for close that protects from double-close */
	throws XmlException {
	if (swigCPtr != 0) {
	    closeInternal();
	    swigCPtr = 0;
	}
    }
    protected void release()
    {
	swigCMemOwn = false;
	swigCPtr = 0;
    }
    %}
    
%typemap(javacode) XmlInputStream %{
    protected void release()
    {
	swigCMemOwn = false;
	swigCPtr = 0;
    }
    %}

%javamethodmodifiers XmlResults::nextInternal() "protected";
%javamethodmodifiers XmlResults::previousInternal() "protected";
%javamethodmodifiers XmlResults::peekInternal() "protected";
%javamethodmodifiers XmlResults::nextDocumentInternal() "protected";
%javamethodmodifiers XmlResults::previousDocumentInternal() "protected";
%javamethodmodifiers XmlResults::peekDocumentInternal() "protected";
%javamethodmodifiers XmlResults::addInternal(const XmlValue&) "protected";
%javamethodmodifiers XmlResults::resetInternal() "protected";
%javamethodmodifiers XmlResults::copy(const XmlResults&) "protected";

%extend XmlResults{
    static XmlResults *copy(const XmlResults &results)
    {
	return new XmlResults(results);
    }
    XmlValue nextInternal() {
	XmlValue value;
	if (self->next(value)){
	    if(value.isNode()){
		if((*value).getDbXmlNodeValueObject()
			 && value.asDocument()
			 && (*value.asDocument()).getDbMinder() != 0
			 && (**self).getDbMinder() == 0){
		    (**self).getDbMinder() = (*value.asDocument()).getDbMinder();
		}
		if(!(*value).getRawNodeValueObject())
		    return new RawNodeValue(value, *self);
	    }
	}
	return value;
    }
    XmlValue previousInternal() {
	XmlValue value;
	if (self->previous(value)){
	    if(value.isNode()){
		if((*value).getDbXmlNodeValueObject()
			 && value.asDocument()
			 && (*value.asDocument()).getDbMinder() != 0
			 && (**self).getDbMinder() == 0){
		    (**self).getDbMinder() = (*value.asDocument()).getDbMinder();
		}
		if(!(*value).getRawNodeValueObject())
		    return new RawNodeValue(value, *self);
	    }
	}
	return value;
    }
    XmlValue peekInternal() {
	XmlValue value;
	if (self->peek(value)){
	    if(value.isNode()){
		if((*value).getDbXmlNodeValueObject()
			 && value.asDocument()
			 && (*value.asDocument()).getDbMinder() != 0
			 && (**self).getDbMinder() == 0){
		    (**self).getDbMinder() = (*value.asDocument()).getDbMinder();
		}
		if(!(*value).getRawNodeValueObject())
		    return new RawNodeValue(value, *self);
	    }
	}
	return value;
    }
    /* If a node is added that is protected by a transaction then it should
       be added to the Results context.  Mulitiple transactions are not allowed. */
    void addInternal(const XmlValue &value)
    {
	const RawNodeValue *raw = (*value).getRawNodeValueObject();
	if(raw){
	    Transaction *txn = raw->getTransaction();
	    if(txn){
		Transaction *resultsTxn = (**self).getOperationContext().txn();
		if(resultsTxn){
		    if(&*txn != &*resultsTxn){
			throw XmlException(XmlException::TRANSACTION_ERROR, "You should not use multiple transactions.");
		    }
		}else
		    (**self).getOperationContext().set(txn); 
	    }
	}
	self->add(value);
    }
    XmlDocument nextDocumentInternal()
    {
	XmlDocument cdoc;
	if(self->next(cdoc))
	    return cdoc;
	return 0;
    }
    XmlDocument previousDocumentInternal()
    {
	XmlDocument cdoc;
	if(self->previous(cdoc))
	    return cdoc;
	return 0;
    }
    XmlDocument peekDocumentInternal()
    {
	XmlDocument cdoc;
	if(self->peek(cdoc))
	    return cdoc;
	return 0;
    }
}

%typemap(javacode) XmlResults %{
    protected HashMap resultsMap = null;
	
    public XmlResults(XmlResults results) throws XmlException {
	this(dbxml_javaJNI.XmlResults_copy(XmlResults.getCPtr(results), results), true);
	if(results.resultsMap != null) resultsMap = new HashMap(results.resultsMap);
    }

    public XmlValue next() throws XmlException
    {
	XmlValue value = nextInternal();
	if(value == null || value.isNull())
	    return null;
	if(value.getReturnResultPtr() != 0)
	    value.setResult(resultsMapGet(new Long(value.getReturnResultPtr())));
	else if(value.isNode())
	    value.setResult(this);
	return value;
    }
    
    public XmlValue previous() throws XmlException
    {
	XmlValue value = previousInternal();
	if(value == null || value.isNull())
	    return null;
	if(value.getReturnResultPtr() != 0)
	    value.setResult(resultsMapGet(new Long(value.getReturnResultPtr())));
	else if(value.isNode())
	    value.setResult(this);
	return value;
    }
    
    public XmlValue peek() throws XmlException
    {
	XmlValue value = peekInternal();
	if(value == null || value.isNull())
	    return null;
	if(value.getReturnResultPtr() != 0)
	    value.setResult(resultsMapGet(new Long(value.getReturnResultPtr())));
	else if(value.isNode())
	    value.setResult(this);
	return value;
    }
    
    public void add(XmlValue value) throws XmlException
    {
	if(value == null || value.isNull())
	    throw new XmlException(XmlException.INVALID_VALUE, 
				   "A null XmlValue object cannot be added to the result set.");
	if(value.getResultPtr() != 0)
	    resultsMapPut(new Long(value.getResultPtr()), value.getResult());
	addInternal(value);
    }
    
    public void reset() throws XmlException
    {
	resetInternal();
    }
    
    public boolean next(XmlDocument document) throws XmlException
    {
	XmlDocument doc = nextDocumentInternal();
	if(doc == null)
	    return false;
	document.copy(doc);
	document.setResults(this);
	return true;
    }
    
    public boolean previous(XmlDocument document) throws XmlException
    {
	XmlDocument doc = previousDocumentInternal();
	if(doc == null)
	    return false;
	document.copy(doc);
	document.setResults(this);
	return true;
    }
    
    public boolean peek(XmlDocument document) throws XmlException
    {
	XmlDocument doc = peekDocumentInternal();
	if(doc == null)
	    return false;
	document.copy(doc);
	document.setResults(this);
	return true;
    }
    
    protected XmlResults resultsMapGet(Long key){
	if(resultsMap == null) return null;
	return (XmlResults)resultsMap.get(key);
    }
    protected void resultsMapPut(Long key, XmlResults value){
	if(resultsMap == null) resultsMap = new HashMap();
	resultsMap.put(key, value);
    }
    %}

/*
 * This is a hack to the typemap from java.swg for construction
 * that makes the writer, and optionally the reader inaccessible
 * after construction of XmlEventReaderToWriter.  It does not
 * appear possible to use %rename on constructors.
 */
%typemap(javaconstruct,directorconnect="\n $moduleJNI.$javaclassname_director_connect(this, swigCPtr, swigCMemOwn, WEAKREF);") XmlEventReaderToWriter { 
    this($imcall, true);$directorconnect
			    if(ownsWriter)
				writer.release();
    if (ownsReader)
	reader.release();
 }
 
%typemap(javacode) XmlEventReaderToWriter %{
    public XmlEventReaderToWriter(XmlEventReader reader, XmlEventWriter writer, boolean ownsReader) 
	throws XmlException { 
	this(reader, writer, ownsReader, true);
    }
    %}

%{
    /* don't use SWIG's array handling - save code space */
    #define SWIG_NOINCLUDE 1

	#if DB_VERSION_MAJOR > 4 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR >= 4)
				   #define DB_EXTRA_EXC 1
				   #endif

				   // XML classes
				   static jclass xml_indexdecl_class;
    static jclass xml_valuedecl_class;
    static jclass xml_docdecl_class;
    static jclass xmlex_class;
    static jclass np_class;
    // DB classes
    static jclass dbenv_class;
    static jclass dbtxn_class;
    // DB exception classes
    static jclass dbex_class, deadex_class, lockex_class, memex_class;
    static jclass runrecex_class, rephandledeadex_class;
    #ifdef DB_EXTRA_EXC
	// DB 4.4 TBD...
	static jclass repdupmasterex_class;
    static jclass repholdelectionex_class, repunavailex_class;
    static jclass versionex_class;
    #endif


	// fields
	jfieldID dbenv_private_fid;
    jfieldID dbtxn_private_fid;

    // method IDs
    static jmethodID xml_indexdecl_construct;
    static jmethodID xml_valuenode_construct;
    static jmethodID xml_valuestr_construct;
    static jmethodID xml_valuebin_construct;
    static jmethodID xml_valuenull_construct;
    static jmethodID xml_doc_construct;
    static jmethodID xmlex_construct;
    // DB exception method ids
    static jmethodID dbex_construct, deadex_construct, lockex_construct;
    static jmethodID memex_construct;
    static jmethodID rephandledeadex_construct;
    static jmethodID runrecex_construct;
    #ifdef DB_EXTRA_EXC
	// DB 4.4. TBD
	static jmethodID repdupmasterex_construct, memex_update_method;
    static jmethodID repholdelectionex_construct, repunavailex_construct;
    static jmethodID versionex_construct;
    #endif

	#define DB_PKG "com/sleepycat/db/"
	#define DB_PKG_INT "com/sleepycat/db/internal/"
	#define XML_PKG "com/sleepycat/dbxml/"

	/* Forward declarations */
	static jthrowable __dbj_get_except(JNIEnv *jenv,
					   int err, const char *msg, jobject obj, jobject jdbenv);

    static jthrowable make_db_except(JNIEnv *jenv,
				     int err, const char *msg, jobject obj, jobject jdbenv);

    // all classes
    const struct {
	jclass *cl;
	const char *name;
    } all_classes[] = {
	// XML
	{ &xml_indexdecl_class, XML_PKG "XmlIndexDeclaration" },
	{ &xml_valuedecl_class, XML_PKG "XmlValue" },
	{ &xml_docdecl_class, XML_PKG "XmlDocument" },
	{ &xmlex_class, XML_PKG "XmlException" },
	// DB
	{ &dbenv_class, DB_PKG_INT "DbEnv" },
	{ &dbtxn_class, DB_PKG_INT "DbTxn" },
	{ &dbex_class, DB_PKG "DatabaseException" },
	{ &deadex_class, DB_PKG "DeadlockException" },
	{ &lockex_class, DB_PKG "LockNotGrantedException" },
	{ &memex_class, DB_PKG "MemoryException" },
	{ &rephandledeadex_class, DB_PKG "ReplicationHandleDeadException" },
	{ &runrecex_class, DB_PKG "RunRecoveryException" },
	#ifdef DB_EXTRA_EXC
	{ &repdupmasterex_class, DB_PKG "ReplicationDuplicateMasterException" },
	{ &repholdelectionex_class, DB_PKG "ReplicationHoldElectionException" },
	{ &repunavailex_class, DB_PKG "ReplicationSiteUnavailableException" },
	{ &versionex_class, DB_PKG "VersionMismatchException" },
	#endif
	// Misc
	{ &np_class, "java/lang/NullPointerException" }
    };

    const struct {
	jmethodID *mid;
	jclass *cl;
	const char *name;
	const char *sig;
    } all_methods[] = {
	// XML methods
	{ &xml_indexdecl_construct, &xml_indexdecl_class, "<init>",
	  "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"  },
	{ &xml_valuenode_construct, &xml_valuedecl_class, "<init>",
	  "(IS[B)V"  },
	{ &xml_valuestr_construct, &xml_valuedecl_class, "<init>",
	  "(ILjava/lang/String;Z)V"  },
	{ &xml_valuebin_construct, &xml_valuedecl_class, "<init>",
	  "(I[B)V"  },
	{ &xml_valuenull_construct, &xml_valuedecl_class, "<init>",
	  "()V"  },
	{ &xml_doc_construct, &xml_docdecl_class, "<init>",
	  "(JI)V"  },
	{ &xmlex_construct, &xmlex_class, "<init>",
	  "(ILjava/lang/String;L" DB_PKG "DatabaseException;III)V"  },
	// DB methods
	{ &dbex_construct, &dbex_class, "<init>",
	  "(Ljava/lang/String;I)V" },
	{ &deadex_construct, &deadex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" },
	{ &lockex_construct, &lockex_class, "<init>",
	  "(Ljava/lang/String;IIL" DB_PKG "DatabaseEntry;L" DB_PKG "internal/DbLock;IL" DB_PKG "internal/DbEnv;)V" },
	{ &memex_construct, &memex_class, "<init>",
	  "(Ljava/lang/String;L" DB_PKG "DatabaseEntry;IL" DB_PKG "internal/DbEnv;)V" },
	{ &rephandledeadex_construct, &rephandledeadex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" },
	{ &runrecex_construct, &runrecex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" }
	#ifdef DB_EXTRA_EXC
	,{ &memex_update_method, &memex_class, "updateDatabaseEntry",
	   "(L" DB_PKG "DatabaseEntry;)V" },
	{ &repdupmasterex_construct, &repdupmasterex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" },
	{ &repholdelectionex_construct, &repholdelectionex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" },
	{ &repunavailex_construct, &repunavailex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" },
	{ &versionex_construct, &versionex_class, "<init>",
	  "(Ljava/lang/String;IL" DB_PKG "internal/DbEnv;)V" }
	#endif
    };

    #define NELEM(x) (sizeof (x) / sizeof (x[0]))

	static bool checkDbVersion()
	{
	    int major, minor, patch, dmajor, dminor, dpatch;
	    (void) db_version(&major, &minor, &patch);
	    (void) DbXml::dbxml_db_version(&dmajor, &dminor, &dpatch);
	    // match major and minor numbers
	    // TBD: warn if patch level is different
	    if (major != dmajor || minor != dminor)
		return false;
	    return true;
	}

    extern "C"
	JNIEXPORT void JNICALL Java_com_sleepycat_dbxml_dbxml_1javaJNI_initialize
	(JNIEnv *jenv, jclass clazz)
	{
	    (void)clazz; /* unused */
	    jclass cl;
	    unsigned int i, j;

	    if (checkDbVersion() == false) {
		fprintf(stderr,
			"Mismatched Berkeley DB libraries");
		return;
	    }
	    // This initialization code comes from DB's java_util.i file
	    for (i = 0; i < NELEM(all_classes); i++) {
		cl = (jenv)->FindClass(all_classes[i].name);
		if (cl == NULL) {
		    fprintf(stderr,
			    "Failed to load class %s - check CLASSPATH\n",
			    all_classes[i].name);
		    return;
		}
		/*
		 * Wrap classes in GlobalRefs so we keep the reference between
		 * calls.
		 */
		*all_classes[i].cl = (jclass)(jenv)->NewGlobalRef(cl);

		if (*all_classes[i].cl == NULL) {
		    fprintf(stderr,
			    "Failed to create a global reference for %s\n",
			    all_classes[i].name);
		    return;
		}
	    }

	    // get some field ids
	    #if DB_VERSION_MAJOR < 4 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR <= 1)
				       dbenv_private_fid = jenv->GetFieldID(dbenv_class, "private_dbobj_", "J");
	    dbtxn_private_fid = jenv->GetFieldID(dbtxn_class, "private_dbobj_", "J");
	    #else
		dbenv_private_fid = jenv->GetFieldID(dbenv_class, "swigCPtr", "J");
	    dbtxn_private_fid = jenv->GetFieldID(dbtxn_class, "swigCPtr", "J");
	    #endif
		/* Get method IDs */
		for (i = 0; i < NELEM(all_methods); i++) {
		    *all_methods[i].mid = (jenv)->
			GetMethodID(*all_methods[i].cl, all_methods[i].name,
				    all_methods[i].sig);

		    if (*all_methods[i].mid == NULL) {
			for (j = 0; j < NELEM(all_classes); j++)
			    if (all_methods[i].cl == all_classes[j].cl)
				break;
			fprintf(stderr,
				"Failed to look up method %s.%s with sig %s\n",
				all_classes[j].name, all_methods[i].name,
				all_methods[i].sig);
			return;
		    }
		}
	}
	
    /* Union to convert longs to pointers (see {get,set}_private_dbobj). */
    typedef union {
	jlong java_long;
	void *ptr;
    } long_to_ptr;

    static DbEnv *get_DbEnv(JNIEnv *jenv, jobject jdbenv)
    {
	if (jdbenv == NULL)
	    return NULL;
	else {
	    #if DB_VERSION_MAJOR < 4 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR <= 1)
				       long_to_ptr lp;
	    lp.java_long = jenv->GetLongField(jdbenv, dbenv_private_fid);
	    return DbEnv::wrap_DB_ENV((DB_ENV *)lp.ptr);
	    #else
		jlong swigCPtr = jenv->GetLongField(jdbenv, dbenv_private_fid);
	    return DbEnv::wrap_DB_ENV(*(DB_ENV **)&swigCPtr);
	    #endif
		}
    }

    static DbTxn *get_DbTxn(JNIEnv *jenv, jobject jtxn)
    {
	if (jtxn == NULL)
	    return NULL;
	else {
	    #if DB_VERSION_MAJOR < 4 || (DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR <= 1)
				       long_to_ptr lp;
	    lp.java_long = jenv->GetLongField(jtxn, dbtxn_private_fid);
	    return DbTxn::wrap_DB_TXN((DB_TXN *)lp.ptr);
	    #else
		jlong swigCPtr = jenv->GetLongField(jtxn, dbtxn_private_fid);
	    return DbTxn::wrap_DB_TXN(*(DB_TXN **)&swigCPtr);
	    #endif
		}
    }

    struct DBT_INFO {
	Dbt dbt;
	jbyte *orig_data;
    };

    %}

%{
    // Deals with XmlException, DbException, and std::exception 
    jthrowable createException(JNIEnv *jenv, std::exception *se)
    {
	XmlException *xe = dynamic_cast<XmlException*>(se);
	if (xe) {
	    // use 0 for dberr; otherwise, DatabaseException::toString
	    // prints redundant information.  The DB error information
	    // is already in the what() string.
	    if (xe->getExceptionCode() == XmlException::NULL_POINTER) {
		// throw null pointer exception
		jenv->ThrowNew(np_class, xe->what());
		return NULL;
	    } else {
		jthrowable dt = 0;
		const char *msg = xe->what();
		XmlException::ExceptionCode ecode =
		    xe->getExceptionCode();
		int qLine = xe->getQueryLine();
		int qCol = xe->getQueryColumn();
		if (ecode == XmlException::DATABASE_ERROR) {
		    dt = __dbj_get_except(jenv, xe->getDbErrno(),
					  msg, NULL, NULL);
		}
		return (jthrowable)jenv->
		    NewObject(xmlex_class, xmlex_construct,
			      ecode, jenv->NewStringUTF(msg), dt, 0, qLine, qCol);
	    }
	}
	DbException *de = dynamic_cast<DbException*>(se);
	if(de) {
	    return make_db_except(jenv, de->get_errno(),
				  de->what(), NULL, NULL);
	}
	std::string emsg = "std::exception thrown from C++ API: ";
	emsg += se->what();
	return (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
					   XmlException::INTERNAL_ERROR,
					   jenv->NewStringUTF(emsg.c_str()), 0, 0, 0, 0);
    }
    %}
// Typemaps
%define JAVA_TYPEMAP(_ctype, _jtype, _jnitype)
    %typemap(jstype) _ctype #_jtype
    %typemap(jtype) _ctype #_jtype
    %typemap(jni) _ctype #_jnitype
    %typemap(javain) _ctype "$javainput"
    %typemap(javaout) _ctype { return $jnicall; }
%enddef

    JAVA_TYPEMAP(u_int32_t, int, jint)
    JAVA_TYPEMAP(size_t, int, jint)

    %typemap(in) bool %{
    $1 = ($input != JNI_FALSE);
    %}

%typemap(out) bool %{
    $result = $1 ? JNI_TRUE : JNI_FALSE;
    %}

%typemap(jstype) XmlUpdateContext & "XmlUpdateContext"
    %typemap(javain) XmlUpdateContext & "XmlUpdateContext.getCPtr($javainput)"
    %typemap(in) XmlUpdateContext &, const XmlUpdateContext %{
    //Use the XmlManager to create an XmlUpdateContext
    XmlManager *mgr = *(XmlManager **)&$input; 
    if(mgr == 0){
	jthrowable t = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						   XmlException::INVALID_VALUE,
						   jenv->NewStringUTF("The XmlManager for XmlUpdateContext has been closed or has not be created."), 0, 0, 0, 0);
	jenv->Throw(t);
    }
    XmlUpdateContext uc = mgr->createUpdateContext();
    $1 = &uc;
    %}
 
/* Both the XmlEventWriter and the XmlDocument need to be returned by
 * the function putDocumentAsEventWriter.  This is done by creating a
 * struct to hold both objects, then translating them into a Java XmlDocument
 * that holds a pointer to the XmlEvent Writer 
 */
JAVA_TYPEMAP(DocumentEventWriter, java.lang.Object, jobject)
    %typemap(out) DocumentEventWriter {
    jobject jdoc = 0;
    if($1.doc.isNull()) $result = jdoc;
    else{
	int cid = (*($1.doc)).getContainerID();
	u_int64_t docid = *(u_int64_t *)((*($1.doc)).getID().rawPtr());
	jdoc = jenv->NewObject(xml_docdecl_class, xml_doc_construct, docid, cid);
	jclass cls = jenv->GetObjectClass(jdoc);
	jmethodID mid = jenv->GetMethodID(cls, "setEventWriter","(J)V");
	jlong writerPtr;
	*(XmlEventWriter **)&writerPtr = $1.writer;
	jenv->CallVoidMethod(jdoc, mid, writerPtr);
	$result = jdoc;
    }
}

JAVA_TYPEMAP(DbEnv *, com.sleepycat.db.internal.DbEnv, jobject)
    %typemap(in) DbEnv * {
    $1 = get_DbEnv(jenv, $input);
}

JAVA_TYPEMAP(DbTxn *, com.sleepycat.db.internal.DbTxn, jobject)
    %typemap(in) DbTxn * {
    $1 = get_DbTxn(jenv, $input);
}

//JAVA_TYPEMAP(char *, byte[], jbyteArray)
/*
 * Fake DBT and XmlData * translation to work around SWIG problems
 * with (void *, size_t) typemaps for Java.
 */
JAVA_TYPEMAP(Dbt &, byte[], jbyteArray)
    JAVA_TYPEMAP(Dbt *, byte[], jbyteArray)

    JAVA_TYPEMAP(XmlData &, byte[], jbyteArray)
    JAVA_TYPEMAP(XmlData *, byte[], jbyteArray)

    %typemap(in) Dbt & (struct DBT_INFO ldbt) %{
    if ($input != NULL) {
        ldbt.dbt.set_data(ldbt.orig_data = jenv->GetByteArrayElements($input, (jboolean *)0));
        ldbt.dbt.set_size(jenv->GetArrayLength($input));
    }
    ldbt.dbt.set_flags(DB_DBT_MALLOC);
    $1 = &ldbt.dbt;
    %}

%typemap(freearg) Dbt & %{
    if ($input != NULL)
        jenv->ReleaseByteArrayElements($input, ldbt$argnum.orig_data, 0);
    %}

%typemap(out) const Dbt * %{
    $result = jenv->NewByteArray($1->get_size());
    jenv->SetByteArrayRegion($result, 0, $1->get_size(), (jbyte *)$1->get_data());
    %}

%typemap(in) XmlData *, XmlData & (XmlData xml_data) {
    if ($input != NULL) {
        xml_data.set_data(jenv->GetByteArrayElements($input, (jboolean *)0));
        xml_data.set_size(jenv->GetArrayLength($input));
    }
    $1 = &xml_data;
 }

%typemap(out) const XmlData *, XmlData * %{
    if ($1) {
	$result = jenv->NewByteArray($1->get_size());
	jenv->SetByteArrayRegion($result, 0, $1->get_size(),
				 (jbyte *)$1->get_data());
	delete $1; // done with new XmlData object
    } else {
	$result = NULL;
    }
    %}

%typemap(freearg) const XmlData *, XmlData & %{
    if ($input != NULL)
        jenv->ReleaseByteArrayElements($input, (jbyte *)xml_data$argnum.get_data(), (jint)0);
    %}

JAVA_TYPEMAP(XmlIndexDeclaration *, XmlIndexDeclaration, jobject)
    %typemap(out) XmlIndexDeclaration * %{
    if ($1 == NULL)
        $result = NULL;
    else {
        $result = jenv->NewObject(xml_indexdecl_class, xml_indexdecl_construct,
				  jenv->NewStringUTF($1->uri.c_str()),
				  jenv->NewStringUTF($1->name.c_str()),
				  jenv->NewStringUTF($1->index.c_str()));
        delete $1;
    }
    %}
JAVA_TYPEMAP(XmlQueryContext *, XmlQueryContext, jobject)
    JAVA_TYPEMAP(XmlQueryContext &, XmlQueryContext, jobject)
    JAVA_TYPEMAP(XmlQueryContext, XmlQueryContext, jobject)

    JAVA_TYPEMAP(XmlDocument *, XmlDocument, jobject)
    JAVA_TYPEMAP(XmlDocument &, XmlDocument, jobject)
    JAVA_TYPEMAP(XmlDocument, XmlDocument, jobject)

    JAVA_TYPEMAP(XmlValue *, XmlValue, jobject)
    JAVA_TYPEMAP(XmlValue &, XmlValue, jobject)
    JAVA_TYPEMAP(XmlValue, XmlValue, jobject)
    %typemap(javadirectorin) XmlValue &reslt "reslt"


/* A very ugly hack used so that XmlResolver.resolveDocument passes the resolved
   document stored in XmlValue to the C++ from the Java.  If the typemap
   directorargout is ever implemented for Java, use that to implement this. */
%typemap(directorin,descriptor="Lcom/sleepycat/dbxml/XmlValue;") (XmlValue &reslt)
%{
$input = jenv->NewObject(xml_valuedecl_class, xml_valuenull_construct);
    jresult = (jboolean) jenv->CallStaticBooleanMethod(Swig::jclass_dbxml_javaJNI, Swig::director_methids[2], jobj, jtxn, jmgr, juri, jreslt);
if (jenv->ExceptionOccurred()) return c_result;
    c_result = jresult ? true : false; 
} else {
    SWIG_JavaThrowException(jenv, SWIG_JavaNullPointerException, "null upcall object");
}
if (jobj) jenv->DeleteLocalRef(jobj);
jclass cls = jenv->GetObjectClass($input);
jmethodID mid = jenv->GetMethodID(cls, "asDocument", "()Lcom/sleepycat/dbxml/XmlDocument;");
jobject jdoc = jenv->CallObjectMethod($input, mid);
if(jdoc){
    XmlDocument cdoc = createCPPXmlDocument(jenv, jdoc);
    XmlValue val(cdoc);
    $1 = val;
}
if($input) jenv->DeleteLocalRef($input);
return c_result;
if(true){
%}

%{
    XmlDocument createCPPXmlDocument(JNIEnv *jenv, jobject jdoc, bool setData = true);
    jobject createJavaXmlDocument(JNIEnv *jenv, XmlDocument &cdoc);

    //Translate a Java XmlValue to a C++ XmlValue when it is passed as an argument
    XmlValue createCPPXmlValue(JNIEnv *jenv, jobject jvalue, bool *valid = 0)
    {
	bool createdResults = false;
	XmlResults xmlresult;
	jlong resultPtr = 0;
	jthrowable thr = NULL;
	try {
		if(jvalue) {
			//Call the Java object to get the type
			jclass cls = jenv->GetObjectClass(jvalue);
			jmethodID mid = jenv->GetMethodID(cls, "getType", "()I");
			jint type = jenv->CallIntMethod(jvalue, mid);
			if(type == XmlValue::NODE){
				/*If it is a node get the data held in the ID
				  and get the XmlResults object either from the
				  Java object or construct it */
				mid = jenv->GetMethodID(cls, "getID", "()[B");
				jbyteArray value = (jbyteArray)jenv->CallObjectMethod(jvalue, mid);
				jbyte *bytes = jenv->GetByteArrayElements(value, (jboolean *)0);
				mid = jenv->GetMethodID(cls, "getResultPtr", "()J");
				resultPtr = (jlong)jenv->CallLongMethod(jvalue, mid);
				if(resultPtr != 0)
					xmlresult = *(*(XmlResults **)&resultPtr);
				else{
					/* If the XmlResults object does not exist in the Java
					   then construct it from the XmlManager */
					createdResults = true;
					mid = jenv->GetMethodID(cls, "getManagerPtr", "()J");
					jlong mgrPtr = (jlong)jenv->CallLongMethod(jvalue, mid);
					if (mgrPtr != 0){
						XmlManager *mgr = *(XmlManager **)&mgrPtr;
						mid = jenv->GetMethodID(cls, "getTransactionPtr", "()J");
						jlong txnPtr = (jlong)jenv->CallLongMethod(jvalue, mid);
						if(txnPtr != 0){
							XmlTransaction *txn = *(XmlTransaction **)&txnPtr;
							xmlresult = new ValueResults(*mgr, *txn);
						}else
							xmlresult = mgr->createResults();
					}else{
						XmlException e(XmlException::INVALID_VALUE, "The result set for this value has been closed, cannot access value.");
						throw e;
					}
				}
				RawNodeValue *raw = new RawNodeValue((void *)bytes, xmlresult);
				jenv->ReleaseByteArrayElements(value, bytes, 0);
				/*If the results where not created for this call then that
				  information should be recorded so that if the value is created
				  again in Java using the XmlResults.next/previous/peek function
				  then it will be set the its original XmlResults Java object */
				if(!createdResults) {
				    XmlResults *res = *(XmlResults **)&resultPtr;
				    raw->setReturnResults(res);
				}
				mid = jenv->GetMethodID(cls, "getDocument", "()Lcom/sleepycat/dbxml/XmlDocument;");
				jobject jdoc = jenv->CallObjectMethod(jvalue, mid);
				if(jdoc){
					XmlDocument cdoc = createCPPXmlDocument(jenv, jdoc);
					raw->setDocument(cdoc);
				}
				return raw;
			}else if(type == XmlValue::BINARY){
				mid = jenv->GetMethodID(cls, "asBinary", "()[B");
				jbyteArray value = (jbyteArray)jenv->CallObjectMethod(jvalue, mid);
				jbyte *bytes = jenv->GetByteArrayElements(value, (jboolean *)0);
				XmlData data((void *)bytes, jenv->GetArrayLength(value));
				XmlValue returnValue((XmlValue::Type)type, data);
				jenv->ReleaseByteArrayElements(value, bytes, 0);
				return returnValue;
			}else if(type == XmlValue::NONE){
				return XmlValue();
			}else{
				mid = jenv->GetMethodID(cls, "asString", "()Ljava/lang/String;");
				jstring value = (jstring)jenv->CallObjectMethod(jvalue, mid);
				const char *str = jenv->GetStringUTFChars(value, (jboolean *)0);
				std::string stringValue(str);
				jenv->ReleaseStringUTFChars(value, str);
				return XmlValue((XmlValue::Type)type, stringValue);
			}
		}
	}catch (std::exception &se) {
		thr = createException(jenv, &se);
	}
	catch(JavaException & /* je */) {
		// This means there's already an exception waiting in the JVM
		XmlQueryContext nullContext;
		if (valid)
			*valid = false;
		return XmlValue();
	}
	catch (...) {
		thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						  XmlException::INTERNAL_ERROR,
						  jenv->NewStringUTF("Uncaught exception from C++ API Constructing XmlValue"), 0, 0, 0, 0);
	}
	if (thr) {
		XmlQueryContext nullContext;
		jenv->Throw(thr);
	}
	if (valid)
		*valid = false;
	return XmlValue(); // empty return
    }
	
    //Translate a C++ XmlValue into a java XmlValue when it is returned from a function
    jobject createJavaXmlValue(JNIEnv *jenv, const XmlValue &xmlvalue)
    {
        jthrowable thr = NULL;
	jobject jvalue = 0;
	try {
		if(xmlvalue.isNode()){
			/* If the XmlValue is a node then create a Java XmlValue using
			   the type, node type, and ID. */
			RawNodeValue *node = const_cast<RawNodeValue*>((*xmlvalue).getRawNodeValueObject());
			if(!node){
				return jvalue;
			}else{
				jbyte *data = 0;
				int size = node->dumpValue((void *&)data);
				jbyteArray dataBytes = jenv->NewByteArray(size);
				jenv->SetByteArrayRegion(dataBytes, 0, jenv->GetArrayLength(dataBytes), data);
				jvalue = jenv->NewObject(xml_valuedecl_class, xml_valuenode_construct,
							 xmlvalue.getType(), 
							 xmlvalue.getNodeType(), 
							 dataBytes);
				/* If the new XmlValue was created from another Java XmlValue that had
				   an XmlResult, then set the new value's results to that of the old values */
				if(node->getReturnResults())
				{
					jclass cls = jenv->GetObjectClass(jvalue);
					jmethodID mid = jenv->GetMethodID(cls, "setResultFromPtr", "(J)V");
					jlong resultsPtr = 0;
					*(XmlResults **)&resultsPtr = node->getReturnResults();
					jenv->CallVoidMethod(jvalue, mid, resultsPtr);
				}
				if(node->getNodeType() == XmlValue::DOCUMENT_NODE){
					XmlDocument cdoc = node->asDocument();
					jclass cls = jenv->GetObjectClass(jvalue);
					jmethodID mid = jenv->GetMethodID(cls, "setDocument", "(Lcom/sleepycat/dbxml/XmlDocument;)V");
					jobject jdoc = createJavaXmlDocument(jenv, cdoc);
					jenv->CallVoidMethod(jvalue, mid, jdoc);
				}
			}
		}else if ( xmlvalue.isBinary() ){
			XmlData value = xmlvalue.asBinary();
			jbyteArray valueBytes = jenv->NewByteArray(value.get_size());
			jbyte *valueData = static_cast<jbyte*>(value.get_data());
			jenv->SetByteArrayRegion(valueBytes, 0, jenv->GetArrayLength(valueBytes), valueData);
			jvalue = jenv->NewObject(xml_valuedecl_class, xml_valuebin_construct,
						 xmlvalue.getType(),
						 valueBytes);
		}else if (xmlvalue.isNull())
			return jenv->NewObject(xml_valuedecl_class, xml_valuenull_construct);
		else{
			//Can be a binary or a string value
			try{
				std::string value = xmlvalue.asString();
				return jenv->NewObject(xml_valuedecl_class, xml_valuestr_construct,
						       xmlvalue.getType(),
						       jenv->NewStringUTF(value.c_str()),
						       false);
			}catch(...){}
			if(!jvalue){
				try{
					XmlData value = xmlvalue.asBinary();
					jbyteArray valueBytes = jenv->NewByteArray(value.get_size());
					jbyte *valueData = static_cast<jbyte*>(value.get_data());
					jenv->SetByteArrayRegion(valueBytes, 0, jenv->GetArrayLength(valueBytes), valueData);
					return jenv->NewObject(xml_valuedecl_class, xml_valuebin_construct,
							       xmlvalue.getType(),
							       valueBytes);
				}catch(...){
					XmlException e(XmlException::INTERNAL_ERROR, "Uncaught exception from C++ API");
					throw e;
				}
			}
		}
	} catch (std::exception &se) {
		thr = createException(jenv, &se);
	}
	catch(JavaException & /* je */) {
		// This means there's already an exception waiting in the JVM
		jvalue = 0;
	}
	catch (...) {
		thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						  XmlException::INTERNAL_ERROR,
						  jenv->NewStringUTF("Uncaught exception from C++ API Constructing Java XmlValue"), 0, 0, 0, 0);
	}
	if (thr) {
		jenv->Throw(thr);
		jvalue = 0;
	}
	return jvalue;
    }
    
    //Set the altered metadata in a Java XmlDocument from a C++ XmlDocument
    void createJavaMetaData(JNIEnv *jenv, jobject jdoc, XmlDocument &cdoc)
    {
	if(!jdoc || cdoc.isNull()) return;
		
	XmlMetaDataIterator itr = cdoc.getMetaDataIterator();
	std::string uri, name;
	XmlValue value;
	jclass cls = jenv->GetObjectClass(jdoc);
	jmethodID mid = jenv->GetMethodID(cls, "addMetaData", 
					  "(Ljava/lang/String;Ljava/lang/String;Lcom/sleepycat/dbxml/XmlValue;ZZ)V");
	while(itr.next(uri, name, value)){
	    if((*itr).previousIsModified()){
		jobject jvalue = createJavaXmlValue(jenv, value);
		jenv->CallVoidMethod(jdoc, mid, jenv->NewStringUTF(uri.c_str()),
				     jenv->NewStringUTF(name.c_str()), jvalue, JNI_TRUE, JNI_FALSE);
	    }
	    else if((*itr).previousIsRemoved()){
		jobject jvalue = createJavaXmlValue(jenv, value);
		jenv->CallVoidMethod(jdoc, mid, jenv->NewStringUTF(uri.c_str()),
				     jenv->NewStringUTF(name.c_str()), jvalue, JNI_FALSE, JNI_TRUE);
	    }
	}
    }
    //Create a C++ XmlQueryContext object from a Java XmlQueryContext object
    XmlQueryContext createCPPXmlQueryContext(JNIEnv *jenv,
					     jobject jqc,
					     jclass cls)
    {
	XmlQueryContext cqc;
	{
	    jthrowable thr = NULL;
	    try {
		if(!jqc) return cqc;
		jmethodID mid = jenv->GetMethodID(cls, "getManagerPtr", "()J");
		jlong mgrPtr = (jlong)jenv->CallLongMethod(jqc, mid);
		if (mgrPtr == 0){
		    XmlException e(XmlException::INVALID_VALUE, "No manager exists for the XmlQueryContext, cannot preceed.");
		    throw e;
		}
		XmlManager *mgr = *(XmlManager **)&mgrPtr;
		//get the return and evaluation types
		jmethodID eMid = jenv->GetMethodID(cls, "getEvaluationType", "()I");
		int et = jenv->CallIntMethod(jqc, eMid);
		cqc = mgr->createQueryContext(XmlQueryContext::LiveValues, (XmlQueryContext::EvaluationType)et);
		jmethodID pack = jenv->GetMethodID(cls, "pack", "()[Ljava/lang/Object;");
		jobjectArray objects = (jobjectArray)jenv->CallObjectMethod(jqc, pack);
		int arrayLength = jenv->GetArrayLength(objects);
		//get the base uri
		int i = 0;
		const char *curi;
		jstring juri = (jstring)jenv->GetObjectArrayElement(objects, i++);
		if(juri != 0){
		    curi = jenv->GetStringUTFChars(juri, (jboolean *)0);
		    std::string uri(curi);
		    jenv->ReleaseStringUTFChars(juri, curi);
		    cqc.setBaseURI(uri);
		}
		//get the default collection
		juri = (jstring)jenv->GetObjectArrayElement(objects, i++);
		if(juri != 0){
		    curi = jenv->GetStringUTFChars(juri, (jboolean *)0);
		    std::string uri2(curi);
		    jenv->ReleaseStringUTFChars(juri, curi);
		    cqc.setDefaultCollection(uri2);
		}
		//get the Query Interrupt time
		jmethodID qMid = jenv->GetMethodID(cls, "getQueryTimeoutSeconds", "()I");
		unsigned int queryTimeout = jenv->CallIntMethod(jqc, qMid);
		cqc.setQueryTimeoutSeconds(queryTimeout);
		//get the namespaces
		jobject jprefix = jenv->GetObjectArrayElement(objects, i++);
		while(arrayLength > i && jprefix != 0){
		    //prefix
		    curi = jenv->GetStringUTFChars((jstring)jprefix, (jboolean *)0);
		    std::string prefix(curi);
		    jenv->ReleaseStringUTFChars((jstring)jprefix, curi);
		    //uri
		    juri = (jstring)jenv->GetObjectArrayElement(objects, i++);
		    curi = jenv->GetStringUTFChars(juri, (jboolean *)0);
		    std::string uri(curi);
		    jenv->ReleaseStringUTFChars(juri, curi);
		    cqc.setNamespace(prefix, uri);
		    jprefix = jenv->GetObjectArrayElement(objects, i++);
		}
		//get variables
		while(arrayLength > i){
		    //name
		    jstring jname = (jstring)jenv->GetObjectArrayElement(objects, i++);
		    curi = jenv->GetStringUTFChars(jname, (jboolean *)0);
		    std::string name(curi);
		    jenv->ReleaseStringUTFChars(jname, curi);
		    //value
		    jobjectArray jvalues = (jobjectArray)jenv->GetObjectArrayElement(objects, i++);
		    int valueArrayLength = jenv->GetArrayLength(jvalues);
		    XmlResults results = mgr->createResults(); 
		    for(int i = 0; i < valueArrayLength; i++){
			jobject jvalue = jenv->GetObjectArrayElement(jvalues, i);
			XmlValue value = createCPPXmlValue(jenv, jvalue);
			results.add(value); 
		    }
		    results.reset();
		    //set the variable
		    cqc.setVariableValue(name, results);
		}
	    } catch (std::exception &se) {
		thr = createException(jenv, &se);
	    }
	    catch(JavaException & /* je */) {
		// This means there's already an exception waiting in the JVM
		    return XmlQueryContext();
	    }
	    catch (...) {
		    thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						      XmlException::INTERNAL_ERROR,
						      jenv->NewStringUTF("Uncaught exception from C++ API Constructing XmlQueryContext"), 0, 0, 0, 0);
	    }

	    if (thr) {
		jenv->Throw(thr);
		return XmlQueryContext();
	    }
	}
	return cqc;
    }
    
    //Set all altered metadata in a C++ XmlDocument from a Java XmlDocument
    void createCPPMetaData(JNIEnv *jenv, jobject jdoc, XmlDocument &cdoc)
    {
	if(!jdoc || cdoc.isNull()) return;
			
	jclass cls = jenv->GetObjectClass(jdoc);
	jmethodID metaDataID = jenv->GetMethodID(cls, "getMetaData", 
						 "(I)Lcom/sleepycat/dbxml/XmlMetaData;");
		
	int index = 0;
	jobject metaData = jenv->CallObjectMethod(jdoc, metaDataID, index);
	if(!metaData)
	    return;
	jclass metaDataCls = jenv->GetObjectClass(metaData);
	jmethodID uriMid = jenv->GetMethodID(metaDataCls, "get_uri", 
					     "()Ljava/lang/String;");
	jmethodID nameMid = jenv->GetMethodID(metaDataCls, "get_name", 
					      "()Ljava/lang/String;");
	jmethodID valueMid = jenv->GetMethodID(metaDataCls, "get_value", 
					       "()Lcom/sleepycat/dbxml/XmlValue;");
	jmethodID modifiedMid = jenv->GetMethodID(metaDataCls, "get_modified", 
						  "()Z");
	jmethodID removedMid = jenv->GetMethodID(metaDataCls, "get_removed", 
						 "()Z");
	while(metaData)
	    {
		/* If the metadata has neither been modifed or removed then it is
		 * stored on disk and does not have to be explicitly added to the
		 * C++ XmlDocument.
		 */
		jboolean jmodifed = jenv->CallBooleanMethod(metaData, modifiedMid);
		jboolean jremoved = jenv->CallBooleanMethod(metaData, removedMid);
		if(jmodifed != JNI_FALSE){
		    jstring juri = (jstring)jenv->CallObjectMethod(metaData, uriMid);
		    const char *curi = jenv->GetStringUTFChars(juri, (jboolean *)0);
		    std::string uri(curi);
		    jenv->ReleaseStringUTFChars(juri, curi);
			    
		    jstring jname = (jstring)jenv->CallObjectMethod(metaData, nameMid);
		    const char *cname = jenv->GetStringUTFChars(jname, (jboolean *)0);
		    std::string name(cname);
		    jenv->ReleaseStringUTFChars(jname, cname);
			    
		    jobject jvalue = jenv->CallObjectMethod(metaData, valueMid);
		    XmlValue value = createCPPXmlValue(jenv, jvalue);
				
		    cdoc.setMetaData(uri, name, value);
		}
		if(jremoved != JNI_FALSE){
		    jstring juri = (jstring)jenv->CallObjectMethod(metaData, uriMid);
		    const char *curi = jenv->GetStringUTFChars(juri, (jboolean *)0);
		    std::string uri(curi);
		    jenv->ReleaseStringUTFChars(juri, curi);
			    
		    jstring jname = (jstring)jenv->CallObjectMethod(metaData, nameMid);
		    const char *cname = jenv->GetStringUTFChars(jname, (jboolean *)0);
		    std::string name(cname);
		    jenv->ReleaseStringUTFChars(jname, cname);
				
		    cdoc.removeMetaData(uri, name);
		}
			
		index++;
		metaData = jenv->CallObjectMethod(jdoc, metaDataID, index);
	    }
    }
    jobject createJavaXmlDocument(JNIEnv *jenv, XmlDocument &cdoc)
    {
	jthrowable thr = NULL;
	jobject jdoc = 0;
        try {
		if(cdoc.isNull()) return jdoc;
		int cid = (*cdoc).getContainerID();
		u_int64_t docid = *(u_int64_t *)((*cdoc).getID().rawPtr());
		jdoc = jenv->NewObject(xml_docdecl_class, xml_doc_construct, docid, cid);
		jclass cls = jenv->GetObjectClass(jdoc);
		createJavaMetaData(jenv, jdoc, cdoc);
		int contentType = (*cdoc).getDefinitiveContent();
		if(contentType != Document::NONE){
			/* If the data has been modified or the document is temporary then
			   the content has to be changed to a DBT so it will not be consumed
			   from the C++ document, which is what would happen if the content
			   was an INPUTSTREAM or EVENTREADER.  */
			if((*cdoc).isContentModified() || docid == 0){
				XmlData content = cdoc.getContent();
				jmethodID mid = jenv->GetMethodID(cls, "setContent", "([B)V");
				jbyteArray bytes = jenv->NewByteArray(content.get_size());
				jbyte *data = static_cast<jbyte*>(content.get_data());
				jenv->SetByteArrayRegion(bytes, 0, jenv->GetArrayLength(bytes), data);
				jenv->CallVoidMethod(jdoc, mid, bytes);
			}
		}
	} catch (std::exception &se) {
		thr = createException(jenv, &se);
	}
	catch(JavaException & /* je */) {
		// This means there's already an exception waiting in the JVM
		jdoc = 0;
	}
	catch (...) {
		thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						  XmlException::INTERNAL_ERROR,
						  jenv->NewStringUTF("Uncaught exception from C++ API Constructing Java XmlDocument"), 0, 0, 0, 0);
	}
	if (thr) {
		jenv->Throw(thr);
		jdoc = 0;
	}
	return jdoc;
    }
	
    XmlDocument createCPPXmlDocument(JNIEnv *jenv, jobject jdoc, bool setData)
    {
	    XmlDocument cdoc;
	    jthrowable thr = NULL;
	    if(jdoc == 0)
		    return cdoc;
	    try {
		    jclass cls = jenv->GetObjectClass(jdoc);	
		    jmethodID mid = jenv->GetMethodID(cls, "getManagerPtr", "()J");
		    jlong mgrPtr = (jlong)jenv->CallLongMethod(jdoc, mid);
		    mid = jenv->GetMethodID(cls, "getResultPtr", "()J");
		    jlong resultPtr = (jlong)jenv->CallLongMethod(jdoc, mid);
		    if (mgrPtr != 0){
			    XmlManager *mgr = *(XmlManager **)&mgrPtr;
			    cdoc = mgr->createDocument();
			    mid = jenv->GetMethodID(cls, "getTransactionPtr", "()J");
			    jlong txnPtr = (jlong)jenv->CallLongMethod(jdoc, mid);
			    if(txnPtr != 0){
				    XmlTransaction *txn = *(XmlTransaction **)&txnPtr;
				    (*cdoc).getOperationContext().set(&**txn);
			    }
		    }else if(resultPtr != 0){
			    XmlResults &res = **(XmlResults **)&resultPtr;
			    XmlManager mgr = (*res).getManager();
			    cdoc = mgr.createDocument();
			    (*cdoc).getOperationContext().set((*res).getOperationContext().txn());
			    (*cdoc).setDbMinder((*res).getDbMinder());
		    }else{
			    XmlException e(XmlException::INVALID_VALUE, "Either the manager that created this document, or the results set that return it, has been deleted, cannot access the document.");
			    throw e;
		    }
		    mid = jenv->GetMethodID(cls, "getId", "()J");
		    u_int64_t docid = jenv->CallLongMethod(jdoc, mid);
		    DocID id(docid);
		    (*cdoc).setID(id);
		    mid = jenv->GetMethodID(cls, "getCid", "()I");
		    int cid = jenv->CallIntMethod(jdoc, mid);
		    (*cdoc).setContainerID(cid);
		    
		    if(!setData)
			    return cdoc;
		    mid = jenv->GetMethodID(cls, "getType", "()I");
		    int type = jenv->CallIntMethod(jdoc, mid);
		    switch(type){
		    case Document::READER:
		    {
			    mid = jenv->GetMethodID(cls, "getEventReaderPtr", 
						    "()J");
			    jlong readerPtr = jenv->CallLongMethod(jdoc, mid);
			    XmlEventReader *reader = *(XmlEventReader **)&readerPtr;
			    (*cdoc).setContentAsEventReader(*reader);
			    break;
		    }
		    case Document::INPUTSTREAM:
		    {
			    mid = jenv->GetMethodID(cls, "getInputStreamPtr", 
						    "()J");
			    jlong streamPtr = jenv->CallLongMethod(jdoc, mid);
			    XmlInputStream *stream = *(XmlInputStream **)&streamPtr;
			    (*cdoc).setContentAsInputStream(&stream);
			    break;
		    }
		    case Document::DBT:
		    {
			    mid = jenv->GetMethodID(cls, "getDBT", "()[B");
			    jbyteArray content = (jbyteArray)jenv->CallObjectMethod(jdoc, mid);
			    jbyte *bytes = jenv->GetByteArrayElements(content, (jboolean *)0);
			    DbtOut *data = new DbtOut((void *)bytes, jenv->GetArrayLength(content));
			    jenv->ReleaseByteArrayElements(content, bytes, 0);
			    (*cdoc).setContentAsDbt(&data);
			    break;
		    }
		    default:
			    if(docid != 0)
				    (*cdoc).setAsNotMaterialized();
		    }
		    
		    createCPPMetaData(jenv, jdoc, cdoc);
		    
		    return cdoc;
	    } catch (std::exception &se) {
		    thr = createException(jenv, &se);
	    }
	    catch(JavaException & /* je */) {
		    // This means there's already an exception waiting in the JVM
		    // Fall through
	    }
	    catch (...) {
		    thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						      XmlException::INTERNAL_ERROR,
						      jenv->NewStringUTF("Uncaught exception from C++ API Constructing XmlDocument"), 0, 0, 0, 0);
	    }
	    if (thr)
		    jenv->Throw(thr);
	    return XmlDocument();
    }
    
    /*
     * Class:     com_sleepycat_dbxml_dbxml_javaJNI
     * Method:    materializeMetaData
     * Signature: (Lcom/sleepycat/dbxml/XmlDocument;)V
     * Copies all the meta data in a stored document to the java document.
     */
    extern "C"
	JNIEXPORT void JNICALL Java_com_sleepycat_dbxml_dbxml_1javaJNI_materializeMetaData
	(JNIEnv *jenv, jclass jcls, jobject jdoc)
	{
	    if(!jdoc) return;
	    jthrowable thr = 0;
	    try{
		XmlDocument cdoc = createCPPXmlDocument(jenv, jdoc, false);
		if ((*(u_int64_t *)((*cdoc).getID().rawPtr()) == 0)) return;
		(*cdoc).setAsNotMaterialized();
		XmlMetaDataIterator itr = cdoc.getMetaDataIterator();
		std::string uri, name;
		XmlValue value;
		jclass cls = jenv->GetObjectClass(jdoc);
		jmethodID mid = jenv->GetMethodID(cls, "addMetaData", 
						  "(Ljava/lang/String;Ljava/lang/String;Lcom/sleepycat/dbxml/XmlValue;ZZ)V");
		while(itr.next(uri, name, value)){
		    jobject jvalue = createJavaXmlValue(jenv, value);
		    jenv->CallVoidMethod(jdoc, mid, jenv->NewStringUTF(uri.c_str()),
					 jenv->NewStringUTF(name.c_str()), jvalue, JNI_FALSE, JNI_FALSE);
		}
	    }catch (std::exception &se) {
		thr = createException(jenv, &se);
	    }
	    catch(JavaException & /* je */) {
		// This means there's already an exception waiting in the JVM
		return;
	    }
	    catch (...) {
		thr = (jthrowable)jenv->NewObject(xmlex_class, xmlex_construct,
						  XmlException::INTERNAL_ERROR,
						  jenv->NewStringUTF("Uncaught exception from C++ API"), 0, 0, 0, 0);
	    }
	    if (thr) {
		jenv->Throw(thr);
		return;
	    }
	}
    /*
     * Class:     com_sleepycat_dbxml_dbxml_javaJNI
     * Method:    interruptQuery
     * Signature: (J)V
     */
    extern "C"
	JNIEXPORT void JNICALL Java_com_sleepycat_dbxml_dbxml_1javaJNI_interruptQuery
	(JNIEnv *jenv, jclass cla, jlong contextPtr)
	{
	    XmlQueryContext context((QueryContext *)contextPtr);
	    if(context.isNull()) return;
	    context.interruptQuery();
	}
    %}

%{
struct queryContextRef {
public:
	queryContextRef(JNIEnv *jenv, jobject obj) :
		env_(jenv), obj_(obj) {
		cls_ = env_->GetObjectClass(obj);
		setCPtr_ = env_->GetMethodID(cls_, "setCPtr", "(J)V");
		qc_ = createCPPXmlQueryContext(env_, obj_, cls_);
		//Set the c pointer in Java so the query can be interrupted
		QueryContext *cptr = (QueryContext *)qc_;
		env_->CallVoidMethod(obj_, setCPtr_, (jlong)cptr);
	}
	~queryContextRef() {
		// clear pointer in Java object
		jthrowable ex = env_->ExceptionOccurred();
		if (ex != NULL)
		    env_->ExceptionClear();
		env_->CallVoidMethod(obj_, setCPtr_, (jlong)0);
		if(ex != NULL)
			env_->Throw(ex);
	}
	XmlQueryContext *getQC() { return &qc_; }
	XmlQueryContext qc_;
	JNIEnv *env_;
	jobject obj_;
	jclass cls_;
	jmethodID setCPtr_;
};
%}
    
%typemap(in) XmlQueryContext &%{
	queryContextRef qcRef(jenv, $input);
	if (qcRef.getQC()->isNull())
		return 0; // exception was thrown
	$1 = qcRef.getQC();
%}

%typemap(in) XmlValue & %{
        bool valid = true;
	XmlValue val$argnum = createCPPXmlValue(jenv, $input, &valid);
	if (!valid) return $null;
	$1 = &val$argnum;
%}

//Turns the C++ XmlValue object return by the function into a Java object
%typemap(out) const XmlValue *,  XmlValue *{
    if($1 == NULL)
	    $result = NULL;
    else
	    $result = createJavaXmlValue(jenv, *$1);		    
}

%typemap(out) XmlValue &, XmlValue{
    $result = createJavaXmlValue(jenv, $1);
}

%typemap(out) const XmlValue &{
    $result = createJavaXmlValue(jenv, *$1);	
}


%typemap(in) XmlDocument & %{
	XmlDocument cdoc$argnum = createCPPXmlDocument(jenv, $input);
	if (cdoc$argnum.isNull()) return $null;
	$1 = &cdoc$argnum;
%}


//Turns the C++ XmlDocument object return by the function into a Java object
%typemap(out) const XmlDocument *,  XmlDocument *{
    if($1 == NULL)
	    $result = NULL;
    else
	    $result = createJavaXmlDocument(jenv, *$1);
}

%typemap(out) XmlDocument {
    $result = createJavaXmlDocument(jenv, $1);
}

%typemap(out) const XmlDocument &, XmlDocument &{
    $result = createJavaXmlDocument(jenv, *$1);
 }

class HelperFunctions;

JAVA_XMLEXCEPTION(getContentAsEventWriter)

//%newobject HelperFunctions::asEventReader(XmlData &, XmlResults *) const;
%newobject HelperFunctions::createResults(XmlTransaction*, XmlManager*) const;
%newobject HelperFunctions::getAttributes(XmlData &, XmlResults *) const;
%newobject HelperFunctions::getContentAsXmlInputStream(XmlDocument &doc) const;
%newobject HelperFunctions::getContentAsEventReader(XmlDocument &doc) const;
%newobject XmlContainer::getNode(XmlTransaction txn, String name, XmlDocumentConfig config);
%newobject XmlContainer::getNode(String name);

%javamethodmodifiers HelperFunctions::getTypeURI(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getTypeName(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::asString(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::asDocument(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::asEventReader(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getNodeHandle(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getNodeName(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getNodeValue(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getNamespaceURI(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getPrefix(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getLocalName(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getParentNode(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getFirstChild(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getLastChild(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getPreviousSibling(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getNextSibling(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getAttributes(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::getOwnerElement(XmlData&, XmlResults*) const "protected";
%javamethodmodifiers HelperFunctions::createDocumentValue(XmlDocument&) "protected";
%javamethodmodifiers HelperFunctions::getAtomicTypeName(int, std::string) const "protected";
%javamethodmodifiers HelperFunctions::getAtomicTypeURI(int, std::string) const "protected";
%javamethodmodifiers HelperFunctions::validateString(int, std::string) const "protected";
%javamethodmodifiers HelperFunctions::getContentAsXmlInputStream(XmlDocument &doc) const "protected";
%javamethodmodifiers HelperFunctions::getContentAsEventReader(XmlDocument &doc) const "protected";
%javamethodmodifiers HelperFunctions::getContentAsEventWriter(XmlDocument &doc, XmlEventWriter &writer) const "protected";
%javamethodmodifiers HelperFunctions::getContentAsString(XmlDocument &doc) const "protected";
%javamethodmodifiers HelperFunctions::getContent(XmlDocument &doc) const "protected";
%javamethodmodifiers HelperFunctions::getMetaData(XmlDocument &doc) const "protected";
%javamethodmodifiers HelperFunctions::verifyDefaultCollection(std::string defaultCollection, std::string baseURI) const "protected";
%javamethodmodifiers HelperFunctions::verifyBaseURI(std::string baseURI) const "protected";

class HelperFunctions
{
    public:
	%extend {
	static std::string getTypeURI(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getTypeURI();
	}
	static std::string getTypeName(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getTypeName();
	}
	static std::string asString(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.asString();
	}
	static XmlDocument asDocument(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.asDocument();
	}
	static XmlEventReader &asEventReader(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.asEventReader();
	}
	static std::string getNodeHandle(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getNodeHandle();
	}
	static std::string getNodeName(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getNodeName();
	}
	static std::string getNodeValue(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getNodeValue();
	}
	static std::string getNamespaceURI(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getNamespaceURI();
	}
	static std::string getPrefix(XmlData &data, XmlResults *result) const{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getPrefix();
	}
	static std::string getLocalName(XmlData &data, XmlResults *result) const
	{
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getLocalName();
	}
	static XmlValue getParentNode(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getParentNode();
	}
	static XmlValue getFirstChild(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getFirstChild();
	}
	static XmlValue getLastChild(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getLastChild();
	}
	static XmlValue getPreviousSibling(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getPreviousSibling();
	}
	static XmlValue getNextSibling(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getNextSibling();
	}
	static XmlResults *getAttributes(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    XmlResults *results = new XmlResults(raw.getAttributes(result));
	    return results;
	}
	static XmlValue getOwnerElement(XmlData &data, XmlResults *result) const {
	    RawNodeValue raw(data.get_data(), *result);
	    return raw.getOwnerElement();
	}
	static XmlValue createDocumentValue(XmlDocument &doc) const {
	    XmlManager mgr = (*doc).getManager();
	    XmlResults result = mgr.createResults();
	    XmlValue value(doc);
	    RawNodeValue *raw = new RawNodeValue(value, result);
	    raw->setDocument(doc);
	    return raw;
	}
	static XmlResults createResults(XmlTransaction *txn, XmlManager *mgr) const{
	    return new ValueResults(*mgr, *txn);
	}
	static std::string getAtomicTypeName(int type, std::string v) const
	{
	    AtomicTypeValue value((XmlValue::Type)type, v);
	    return value.getTypeName();
	}
	static std::string getAtomicTypeURI(int type, std::string v) const
	{
	    AtomicTypeValue value((XmlValue::Type)type, v);
	    return value.getTypeURI();
	}
	static void validateString(int type, std::string v) const
	{
	    XmlValue value((XmlValue::Type)type, v);
	}
	/* If the stream is created from a dbt then the data it is reading will be
	   deleted as soon as the function returns to the Java.  If it is read
	   from an event reader then it will remain valid until the stream is closed.
	*/
	static XmlInputStream *getContentAsXmlInputStream(XmlDocument &doc) const
	{
	    XmlEventReader &reader = doc.getContentAsEventReader();
	    doc.setContentAsEventReader(reader);
	    return doc.getContentAsXmlInputStream();
	}
	static XmlEventReader &getContentAsEventReader(XmlDocument &doc) const
	{
	    return doc.getContentAsEventReader();
	}
	static void getContentAsEventWriter(XmlDocument &doc, XmlEventWriter &writer) const
	{
	    doc.getContentAsEventWriter(writer);
	}
	static std::string getContentAsString(XmlDocument &doc) const
	{
	    std::string content;
	    doc.getContent(content);
	    return content;
	}
	static XmlData *getContent(XmlDocument &doc) const
	{
	    return new XmlData(doc.getContent());
	}
	static XmlValue getMetaData(std::string uri, std::string name, XmlDocument &doc) const
	{
	    XmlValue value;
	    doc.getMetaData(uri, name, value);
	    return value;
	}
	static std::string verifyDefaultCollection(XmlManager mgr, std::string defaultCollection, std::string baseURI) const
	{
	    XmlQueryContext context = mgr.createQueryContext();
	    if(baseURI.size() > 0)
		context.setBaseURI(baseURI);
	    context.setDefaultCollection(defaultCollection);
	    return context.getDefaultCollection();
	}
	static void verifyBaseURI(XmlManager mgr, std::string baseURI) const
	{
	    XmlQueryContext context = mgr.createQueryContext();
	    context.setBaseURI(baseURI);
	}
    }/*extend*/
}; /*HelperFunctions*/

// _dbj_* are directly from DB's java_except.i file

%{
    static jthrowable make_db_except(JNIEnv *jenv,
				     int err, const char *msg, jobject obj,
				     jobject jdbenv) {

	jthrowable dt = __dbj_get_except(jenv, err, msg, obj, jdbenv);
	return (jthrowable)jenv->
	    NewObject(xmlex_class, xmlex_construct,
		      XmlException::DATABASE_ERROR,
		      jenv->NewStringUTF(msg), dt, 0, 0, 0, 0);
    }

    static jthrowable __dbj_get_except(
				       JNIEnv *jenv,
				       int err, const char *msg, jobject obj, jobject jdbenv) {
	jobject jmsg;

	if (msg == NULL)
	    msg = db_strerror(err);

	jmsg = jenv->NewStringUTF(msg);

	switch (err) {
	    #if 0
		// don't map these (yet?)
	case EINVAL:
	    return (jthrowable)jenv->
		NewObject(
			  illegalargex_class, illegalargex_construct, jmsg);

	case ENOENT:
	    return (jthrowable)jenv->
		NewObject(filenotfoundex_class, filenotfoundex_construct, jmsg);

	case ENOMEM:
	    return (jthrowable)jenv->
		NewObject(outofmemerr_class, outofmemerr_construct, jmsg);
	    #endif
	case DB_BUFFER_SMALL:
	    return (jthrowable)jenv->
		NewObject(memex_class, memex_construct, jmsg, obj, err, jdbenv);

	case DB_REP_HANDLE_DEAD:
	    return (jthrowable)jenv->
		NewObject(rephandledeadex_class, rephandledeadex_construct,
			  jmsg, err, jdbenv);
	case DB_RUNRECOVERY:
	    return (jthrowable)jenv->
		NewObject(runrecex_class, runrecex_construct, jmsg, err, jdbenv);

	case DB_LOCK_DEADLOCK:
	    return (jthrowable)jenv->
		NewObject(deadex_class, deadex_construct, jmsg, err, jdbenv);

	case DB_LOCK_NOTGRANTED:
	    return (jthrowable)jenv->
		NewObject(lockex_class,lockex_construct,
			  jmsg, 0, 0, NULL, NULL, 0, jdbenv);

	    #ifdef DB_EXTRA_EXC
		// These are from DB 4.4, TBD...
	case DB_REP_DUPMASTER:
	    return (jthrowable)jenv->
		NewObject(repdupmasterex_class, repdupmasterex_construct,
			  jmsg, err, jdbenv);

	case DB_REP_HOLDELECTION:
	    return (jthrowable)jenv->
		NewObject(repholdelectionex_class, repholdelectionex_construct,
			  jmsg, err, jdbenv);

	case DB_REP_UNAVAIL:
	    return (jthrowable)jenv->
		NewObject(repunavailex_class, repunavailex_construct,
			  jmsg, err, jdbenv);

	case DB_VERSION_MISMATCH:
	    return (jthrowable)jenv->
		NewObject(versionex_class,
			  versionex_construct, jmsg, 0, 0, NULL, NULL, 0, jdbenv);
	    #endif //  DB 4.4
	default:
	    return (jthrowable)jenv->
		NewObject(dbex_class,
			  dbex_construct, jmsg, err, jdbenv);
	}
    }

    %}
